import pytest
import numpy as np
from numpy.testing import assert_array_equal

from landlab.grid import raster_funcs as rfuncs
from landlab import RasterModelGrid
from landlab.testing.tools import assert_array_is_int


def test_with_scalars():
    """Test scalar args."""
    rmg = RasterModelGrid(4, 5)
    id = rfuncs.find_nearest_node(rmg, (0.2, 0.6))
    assert id == 5
    assert id.ndim == 0
    assert_array_is_int(id)


def test_with_iterable():
    """Test iterable args."""
    rmg = RasterModelGrid(4, 5)
    id = rfuncs.find_nearest_node(rmg, ([0.2], [0.6]))
    assert_array_equal(id, np.array([5], dtype=int))
    assert_array_is_int(id)


def test_with_ndarray_with_length_0():
    """Test with 0d numpy arrays as args."""
    rmg = RasterModelGrid(4, 5)
    id = rfuncs.find_nearest_node(rmg, (np.array(0.2), np.array(0.6)))
    assert_array_equal(id, np.array(5, dtype=int))
    assert id.ndim == 0
    assert_array_is_int(id)


def test_with_ndarray():
    """Test with 1d numpy arrays as args."""
    rmg = RasterModelGrid(4, 5)
    coords = (np.array([0.1, .2]), np.array([3.4, 2.6]))
    id = rfuncs.find_nearest_node(rmg, coords)
    assert_array_equal(id, np.array([15, 15], dtype=int))
    assert isinstance(id, np.ndarray)


def test_non_unit_spacing():
    """Test with a grid of non-unit spacing."""
    rmg = RasterModelGrid((4, 5), spacing=(2., 2.))
    id = rfuncs.find_nearest_node(rmg, (.9, .2))
    assert id == 0


def test_beyond_grid():
    """Raise an error if points are outside the bounds of the grid."""
    rmg = RasterModelGrid((4, 5), spacing=(2., 2.))

    assert rfuncs.find_nearest_node(rmg, (-.999, .2)) == 0
    with pytest.raises(ValueError):
        rfuncs.find_nearest_node(rmg, (-1.001, .2))

    assert rfuncs.find_nearest_node(rmg, (8.999, .2)) == 4
    with pytest.raises(ValueError):
        rfuncs.find_nearest_node(rmg, (9.001, .2))

    assert rfuncs.find_nearest_node(rmg, (.2, -.999)) == 0
    with pytest.raises(ValueError):
        rfuncs.find_nearest_node(rmg, (.2, -1.001))

    assert rfuncs.find_nearest_node(rmg, (.2, 6.999)) == 15
    with pytest.raises(ValueError):
        rfuncs.find_nearest_node(rmg, (.2, 7.001))




from landlab import Component
from ...utils.decorators import use_file_name_or_kwds
import numpy as np


_VALID_METHODS = set(['Constant', 'PriestleyTaylor', 'MeasuredRadiationPT',
                      'Cosine'])


def _assert_method_is_valid(method):
    if method not in _VALID_METHODS:
        raise ValueError('%s: Invalid method name' % method)


class PotentialEvapotranspiration(Component):

    """
    Potential Evapotranspiration Component calculates spatially distributed
    potential evapotranspiration based on input radiation factor (spatial
    distribution of incoming radiation) using chosen method such as constant
    or Priestley Taylor. Ref: Xiaochi et. al. 2013 for 'Cosine' method and
    ASCE-EWRI Task Committee Report Jan 2005 for 'PriestleyTaylor' method.
    Note: Calling 'PriestleyTaylor' method would generate/overwrite shortwave &
    longwave radiation fields.

    .. codeauthor:: Sai Nudurupati and Erkan Istanbulluoglu

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> from landlab.components.pet import PotentialEvapotranspiration

    >>> grid = RasterModelGrid((5, 4), spacing=(0.2, 0.2))
    >>> PET = PotentialEvapotranspiration(grid)
    >>> PET.name
    'Potential Evapotranspiration'
    >>> PET.input_var_names
    ('radiation__ratio_to_flat_surface',)
    >>> sorted(PET.output_var_names)
    ['radiation__incoming_shortwave_flux',
     'radiation__net_flux',
     'radiation__net_longwave_flux',
     'radiation__net_shortwave_flux',
     'surface__potential_evapotranspiration_rate']
    >>> sorted(PET.units) # doctest: +NORMALIZE_WHITESPACE
    [('radiation__incoming_shortwave_flux', 'W/m^2'),
     ('radiation__net_flux', 'W/m^2'),
     ('radiation__net_longwave_flux', 'W/m^2'),
     ('radiation__net_shortwave_flux', 'W/m^2'),
     ('radiation__ratio_to_flat_surface', 'None'),
     ('surface__potential_evapotranspiration_rate', 'mm')]
    >>> PET.grid.number_of_cell_rows
    3
    >>> PET.grid.number_of_cell_columns
    2
    >>> PET.grid is grid
    True
    >>> pet_rate = grid.at_cell['surface__potential_evapotranspiration_rate']
    >>> np.allclose(pet_rate, 0.)
    True
    >>> grid['cell']['radiation__ratio_to_flat_surface'] = np.array([
    ...       0.38488566, 0.38488566,
    ...       0.33309785, 0.33309785,
    ...       0.37381705, 0.37381705])
    >>> current_time = 0.5
    >>> PET.update(current_time)
    >>> np.allclose(pet_rate, 0.)
    False
    """

    _name = 'Potential Evapotranspiration'

    _input_var_names = (
        'radiation__ratio_to_flat_surface',
    )

    _output_var_names = (
        'surface__potential_evapotranspiration_rate',
        'radiation__incoming_shortwave_flux',
        'radiation__net_shortwave_flux',
        'radiation__net_longwave_flux',
        'radiation__net_flux',
    )

    _var_units = {
        'radiation__ratio_to_flat_surface': 'None',
        'surface__potential_evapotranspiration_rate': 'mm',
        'radiation__incoming_shortwave_flux': 'W/m^2',
        'radiation__net_shortwave_flux': 'W/m^2',
        'radiation__net_longwave_flux': 'W/m^2',
        'radiation__net_flux': 'W/m^2',
    }

    _var_mapping = {
        'radiation__ratio_to_flat_surface': 'cell',
        'surface__potential_evapotranspiration_rate': 'cell',
        'radiation__incoming_shortwave_flux': 'cell',
        'radiation__net_shortwave_flux': 'cell',
        'radiation__net_longwave_flux': 'cell',
        'radiation__net_flux': 'cell',
    }

    _var_doc = {
        'radiation__ratio_to_flat_surface':
            'ratio of total incident shortwave radiation on sloped surface \
             to flat surface',
        'surface__potential_evapotranspiration_rate':
            'potential sum of evaporation and potential transpiration',
        'radiation__incoming_shortwave_flux':
            'total incident shortwave radiation over the time step',
        'radiation__net_shortwave_flux':
            'net incident shortwave radiation over the time step',
        'radiation__net_longwave_flux':
            'net incident longwave radiation over the time step',
        'radiation__net_flux':
            'net total radiation over the time step',
    }

    @use_file_name_or_kwds
    def __init__(self, grid, method='Cosine', priestley_taylor_const=1.26,
                 albedo=0.6, latent_heat_of_vaporization=28.34,
                 psychometric_const=0.066, stefan_boltzmann_const=0.0000000567,
                 solar_const=1366.67, latitude=34.,
                 elevation_of_measurement=300, adjustment_coeff=0.18,
                 lt=0., nd=365., MeanTmaxF=12., delta_d=5., **kwds):
        """
        Parameters
        ----------
        grid: RasterModelGrid
            A grid.
        method: {'Constant', 'PriestleyTaylor', 'MeasuredRadiationPT', 'Cosine'}, optional
            Priestley Taylor method will spit out radiation outputs too.
        priestley_taylor_constant: float, optional
            Alpha used in Priestley Taylor method.
        albedo: float, optional
            Albedo.
        latent_heat_of_vaporization: float, optional
            Latent heat of vaporization for water Pwhv (Wd/(m*mm^2)).
        psychometric_const: float, optional
            Psychometric constant (kPa (deg C)^-1).
        stefan_boltzmann_const: float, optional
            Stefan Boltzmann's constant (W/(m^2K^-4)).
        solar_const: float, optional
            Solar constant (W/m^2).
        latitude: float, optional
            Latitude (radians).
        elevation_of_measurement: float, optional
            Elevation at which measurement was taken (m).
        adjustment_coeff: float, optional
            adjustment coeff to predict Rs from air temperature (deg C)^-0.5.
        lt: float, optional
            lag between peak TmaxF and solar forcing (days).
        nd: float, optional
            Number of days in year (days).
        MeanTmaxF: float, optional
            Mean annual rate of TmaxF (mm/d).
        delta_d: float, optional
            Calibrated difference between max & min daily TmaxF (mm/d).
        """

        self._method = method
        # For Priestley Taylor
        self._alpha = priestley_taylor_const
        self._a = albedo
        self._pwhv = latent_heat_of_vaporization
        self._y = psychometric_const
        self._sigma = stefan_boltzmann_const
        self._Gsc = solar_const
        self._phi = (np.pi / 180.) * latitude
        self._z = elevation_of_measurement
        self._Krs = adjustment_coeff
        self._LT = lt
        self._ND = nd
        self._TmaxF_mean = MeanTmaxF
        self._DeltaD = delta_d
        _assert_method_is_valid(self._method)

        super(PotentialEvapotranspiration, self).__init__(grid, **kwds)

        for name in self._input_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        for name in self._output_var_names:
            if name not in self.grid.at_cell:
                self.grid.add_zeros('cell', name, units=self._var_units[name])

        self._cell_values = self.grid['cell']

    def update(self, current_time=None, const_potential_evapotranspiration=12.,
               Tmin=0., Tmax=1., Tavg=0.5, obs_radiation=350., **kwds):
        """Update fields with current conditions.

        Parameters
        ----------
        current_time: float, required only for 'Cosine' method
            Current time (Years)
        constant_potential_evapotranspiration: float, optional for
            'Constant' method
            Constant PET value to be spatially distributed.
        Tmin: float, required for 'Priestley Taylor' method
            Minimum temperature of the day (deg C)
        Tmax: float, required for 'Priestley Taylor' method
            Maximum temperature of the day (deg C)
        Tavg: float, required for 'Priestley Taylor' and 'MeasuredRadiationPT'
            methods
            Average temperature of the day (deg C)
        obs_radiation float, required for 'MeasuredRadiationPT' method
            Observed radiation (W/m^2)
        """

        if self._method == 'Constant':
            self._PET_value = const_potential_evapotranspiration
        elif self._method == 'PriestleyTaylor':
            self._PET_value = (
                self._PriestleyTaylor(current_time, Tmax, Tmin, Tavg))
            self._cell_values['radiation__incoming_shortwave_flux'] = (
                self._Rs *
                self._cell_values['radiation__ratio_to_flat_surface'])
            self._cell_values['radiation__net_shortwave_flux'] = (
                self._Rns *
                self._cell_values['radiation__ratio_to_flat_surface'])
            self._cell_values['radiation__net_longwave_flux'] = (
                self._Rnl *
                self._cell_values['radiation__ratio_to_flat_surface'])
            self._cell_values['radiation__net_flux'] = (
                self._Rn *
                self._cell_values['radiation__ratio_to_flat_surface'])
        elif self._method == 'MeasuredRadiationPT':
            Robs = obs_radiation
            self._PET_value = self._MeasuredRadPT(Tavg, (1-self._a)*Robs)
        elif self._method == 'Cosine':
            self._J = np.floor((current_time - np.floor(current_time)) * 365.)
            self._PET_value = (
                max((self._TmaxF_mean + self._DeltaD / 2. *
                     np.cos((2 * np.pi) * (self._J - self._LT - self._ND / 2) /
                            self._ND)), 0.0))

        self._PET = (
            self._PET_value *
            self._cell_values['radiation__ratio_to_flat_surface'])
        self._cell_values['surface__potential_evapotranspiration_rate'][:] = (
            self._PET)

    def _PriestleyTaylor(self, current_time, Tmax, Tmin, Tavg):

        # Julian Day - ASCE-EWRI Task Committee Report, Jan-2005 - Eqn 25, (52)
        self._J = np.floor((current_time - np.floor(current_time)) * 365)
        # Saturation Vapor Pressure - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 6, (37)
        self._es = 0.6108 * np.exp((17.27 * Tavg)/(237.7 + Tavg))

        # Actual Vapor Pressure - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 8, (38)
        self._ea = 0.6108 * np.exp((17.27 * Tmin)/(237.7 + Tmin))

        # Slope of Saturation Vapor Pressure - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 5, (36)
        self._delta = (4098.0 * self._es)/((237.3 + Tavg) ** 2.0)

        # Solar Declination Angle - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 24,(51)
        self._sdecl = 0.409 * np.sin(((np.pi / 180.0) * self._J) - 1.39)

        # Inverse Relative Distance Factor - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 23,(50)
        self._dr = 1 + (0.033 * np.cos(np.pi / 180.0 * self._J))

        # To calculate ws - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 29,(61)
        self._x = 1.0 - (((np.tan(self._phi)) ** 2.0) *
                         (np.tan(self._sdecl) ** 2.0))
        if self._x <= 0:
            self._x = 0.00001
            # Sunset Hour Angle - ASCE-EWRI Task Committee Report,
            # Jan-2005 - Eqn 28,(60)
        self._ws = ((np.pi / 2.0) -
                    np.arctan((- 1 * np.tan(self._phi) *
                               np.tan(self._sdecl)) / (self._x ** 2.0)))

        # Extraterrestrial radmodel.docx - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 21, (48)
        # 11.57 converts 1 MJ/m^2/day to W/m^2
        self._Ra = (
                    11.57 * (24.0 / np.pi) * 4.92 * self._dr *
                    ((self._ws * np.sin(self._phi) * np.sin(self._sdecl)) +
                     (np.cos(self._phi) * np.cos(self._sdecl) *
                     (np.sin(self._ws)))))

        # Clear-sky Solar Radiation - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 19, (47)
        self._Rso = (0.75 + ((2.0 * (10 ** (- 5.0))) * self._z)) * self._Ra
        self._Rs = min(self._Krs * self._Ra * np.sqrt(Tmax - Tmin), self._Rso)

        # Net Short Wave Radiation - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 16, (43)
        self._Rns = self._Rs * (1 - self._a)

        # Relative Cloudiness - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Page 20,35
        if self._Rso > 0:
            self._u = self._Rs / self._Rso
        else:
            self._u = 0

        if self._u < 0.3:
            self._u = 0.3
        elif self._u > 1:
            self._u = 1.0

        # Cloudiness Function - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 18, (45)
        self._fcd = (1.35 * self._u) - 0.35

        # Net Long Wave Radiation - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 17, (44)
        self._Rnl = (self._sigma *
                     self._fcd * (0.34 - (0.14 * np.sqrt(self._ea)) *
                                  (((Tmax + 273.16) ** 4.0 +
                                    (Tmin + 273.16) ** 4.0) / 2.0)))

        # Net Radiation - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 15, (42)
        self._Rn = self._Rns - self._Rnl

        self._ETp = max(self._alpha * (self._delta / (self._delta + self._y)) *
                        (self._Rn / self._pwhv), 0)

        return self._ETp

    def _MeasuredRadPT(self, Tavg, Rnobs):
        # Saturation Vapor Pressure - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 6, (37)
        self._es = 0.6108 * np.exp((17.27 * Tavg)/(237.7 + Tavg))

        # Slope of Saturation Vapor Pressure - ASCE-EWRI Task Committee Report,
        # Jan-2005 - Eqn 5, (36)
        self._delta = (4098.0 * self._es) / ((237.3 + Tavg) ** 2.0)
        self._ETp = max(self._alpha * (self._delta / (self._delta + self._y)) *
                        (Rnobs / self._pwhv), 0)
        return self._ETp



# -*- coding: utf-8 -*-
"""
test_sink_filler:

Created on Tues Oct 20, 2015

@author: dejh
"""
import pytest

from numpy import sin, pi
import numpy as np  # for use of np.round
from numpy.testing import assert_array_equal, assert_array_almost_equal

import landlab
from landlab import BAD_INDEX_VALUE as XX
from landlab import RasterModelGrid, FieldError
from landlab.components.flow_routing import FlowRouter
from landlab.components.sink_fill import SinkFiller


def check_fields(sink_grid1):
    """
    Check to make sure the right fields have been created.
    """
    assert_array_equal(
        np.zeros(sink_grid1.number_of_nodes), sink_grid1.at_node["sediment_fill__depth"]
    )
    with pytest.raises(FieldError):
        sink_grid1.at_node["drainage_area"]


def test_get_lake_ext_margin(sink_grid1):
    hf = SinkFiller(sink_grid1)

    lake = np.array([16, 17, 23, 24, 25, 30, 31, 32])
    ext_margin_returned = hf._get_lake_ext_margin(lake)
    ext_margin = np.array(
        [8, 9, 10, 11, 15, 18, 19, 22, 26, 29, 33, 36, 37, 38, 39, 40]
    )
    assert_array_equal(ext_margin_returned, ext_margin)


def test_get_lake_int_margin(sink_grid1):
    hf = SinkFiller(sink_grid1)

    lake = np.array([16, 17, 18, 23, 24, 25, 26, 30, 31, 32])
    ext_margin = np.array(
        [8, 9, 10, 11, 12, 15, 19, 20, 22, 27, 29, 33, 34, 36, 37, 38, 39, 40]
    )
    int_margin_returned = hf._get_lake_int_margin(lake, ext_margin)
    int_margin = np.array([16, 17, 18, 23, 25, 26, 30, 31, 32])
    assert_array_equal(int_margin_returned, int_margin)


def test_drainage_directions_change(sink_grid1):
    hf = SinkFiller(sink_grid1)

    lake = np.array([22, 23])
    old_elevs = np.ones(49, dtype=float)
    old_elevs[lake] = 0.
    new_elevs = old_elevs.copy()
    new_elevs[40] = 2.
    cond = hf.drainage_directions_change(lake, old_elevs, new_elevs)
    assert not cond
    new_elevs[23] = 0.5
    cond = hf.drainage_directions_change(lake, old_elevs, new_elevs)
    assert not cond
    new_elevs[23] = 1.
    cond = hf.drainage_directions_change(lake, old_elevs, new_elevs)
    assert not cond
    new_elevs[23] = 1.2
    cond = hf.drainage_directions_change(lake, old_elevs, new_elevs)
    assert cond


def test_add_slopes(sink_grid1):
    z = sink_grid1.at_node["topographic__elevation"]
    hf = SinkFiller(sink_grid1)

    new_z = z.copy()
    outlet_elev = z[sink_grid1.outlet]
    hf._elev[sink_grid1.lake] = outlet_elev
    rt2 = np.sqrt(2.)
    slope_to_add = 0.1
    lake_map = np.empty_like(z)
    lake_map.fill(XX)
    lake_map[sink_grid1.lake] = sink_grid1.lake_code
    hf._lf._lake_map = lake_map
    hf.lake_nodes_treated = np.array([], dtype=int)
    dists = sink_grid1.calc_distances_of_nodes_to_point(
        (sink_grid1.node_x[sink_grid1.outlet], sink_grid1.node_y[sink_grid1.outlet])
    )
    new_z[sink_grid1.lake] = outlet_elev
    new_z[sink_grid1.lake] += dists[sink_grid1.lake] * slope_to_add
    # test the ones we can do easily analytically separately
    straight_north = np.array([23, 16])
    off_angle = 24
    elevs_out, lake_out = hf._add_slopes(
        slope_to_add, sink_grid1.outlet, sink_grid1.lake_code
    )
    assert_array_equal(
        slope_to_add * (np.arange(2.) + 1.) + outlet_elev, elevs_out[straight_north]
    )
    assert slope_to_add * rt2 + outlet_elev == pytest.approx(elevs_out[off_angle])
    assert_array_equal(new_z, elevs_out)
    assert_array_equal(sink_grid1.lake, lake_out)


def test_filler_flat(sink_grid2):
    """
    Very simple, though possibly degerate, case, filling a 3x3 hole up to
    the flat surface surrounding it.
    """
    hf = SinkFiller(sink_grid2)
    hf.fill_pits()
    assert_array_equal(hf._elev[sink_grid2.lake], np.ones(9, dtype=float))
    assert_array_equal(
        sink_grid2.at_node["topographic__elevation"][sink_grid2.lake],
        np.ones(9, dtype=float),
    )


def test_filler_inclined(sink_grid3):
    """
    Tests a flat fill into an inclined surface, with two holes.
    """
    hf = SinkFiller(sink_grid3)
    hf.fill_pits()
    assert_array_equal(
        sink_grid3.at_node["topographic__elevation"][sink_grid3.lake1],
        np.ones(9, dtype=float) * 4.,
    )
    assert_array_equal(
        sink_grid3.at_node["topographic__elevation"][sink_grid3.lake2],
        np.ones(4, dtype=float) * 7.,
    )


def test_filler_inclined2(sink_grid3):
    """
    Tests an inclined fill into an inclined surface, with two holes.
    """
    z_init = sink_grid3.at_node["topographic__elevation"].copy()
    fr = FlowRouter(sink_grid3)
    hf = SinkFiller(sink_grid3, apply_slope=True)

    hf.fill_pits()
    hole1 = np.array(
        [
            4.00009091,
            4.00018182,
            4.00027273,
            4.00036364,
            4.00045455,
            4.00054545,
            4.00063636,
            4.00072727,
            4.00081818,
        ]
    )
    hole2 = np.array([7.16666667, 7.33333333, 7.5, 7.66666667])
    assert_array_almost_equal(
        sink_grid3.at_node["topographic__elevation"][sink_grid3.lake1], hole1
    )
    assert_array_almost_equal(
        sink_grid3.at_node["topographic__elevation"][sink_grid3.lake2], hole2
    )
    fr.route_flow()
    assert sink_grid3.at_node["flow__sink_flag"][sink_grid3.core_nodes].sum() == 0


def test_stupid_shaped_hole(sink_grid4):
    """Tests inclined fill into a surface with a deliberately awkward shape."""
    fr = FlowRouter(sink_grid4)
    hf = SinkFiller(sink_grid4, apply_slope=True)
    hf.fill_pits()
    hole1 = np.array(
        [
            4.00007692,
            4.00015385,
            4.00023077,
            4.00030769,
            4.00038462,
            4.00046154,
            4.00053846,
            4.00061538,
            4.00069231,
            4.00076923,
            4.00084615,
        ]
    )
    hole2 = np.array([7.4, 7.2, 7.6])

    assert_array_almost_equal(
        sink_grid4.at_node["topographic__elevation"][sink_grid4.lake1], hole1
    )
    assert_array_almost_equal(
        sink_grid4.at_node["topographic__elevation"][sink_grid4.lake2], hole2
    )
    fr.route_flow()
    assert sink_grid4.at_node["flow__sink_flag"][sink_grid4.core_nodes].sum() == 0


def test_D4_routing(sink_grid5):
    """
    Tests inclined fill into a surface with a deliberately awkward shape.
    This is testing D4 routing.
    """
    fr = FlowRouter(sink_grid5)
    hf = SinkFiller(sink_grid5, routing="D4", apply_slope=True)
    hf.fill_pits()
    #    hole1 = np.array([4.00016667, 4.00025, 4.00033333, 4.00008333, 4.00041667,
    #                      4.0005, 4.00066667, 4.00058333, 4.00075,
    #                      4.334])
    hole1 = np.array(
        [
            4.00016667,
            4.00033333,
            4.0005,
            4.00008333,
            4.00025,
            4.00041667,
            4.000833,
            4.00066667,
            4.00058333,
            4.00075,
            4.334,
        ]
    )
    hole2 = np.array([7.6, 7.2, 7.4])

    assert_array_almost_equal(
        sink_grid5.at_node["topographic__elevation"][sink_grid5.lake1], hole1
    )
    assert_array_almost_equal(
        sink_grid5.at_node["topographic__elevation"][sink_grid5.lake2], hole2
    )
    fr.route_flow(method="D4")
    assert sink_grid5.at_node["flow__sink_flag"][sink_grid5.core_nodes].sum() == 0


def test_D4_filling(sink_grid5):
    """
    Tests inclined fill into a surface with a deliberately awkward shape.
    This is testing D4 without inclining the surface.
    """
    hf = SinkFiller(sink_grid5, routing="D4")
    hf.fill_pits()
    hole1 = 4. * np.ones_like(sink_grid5.lake1, dtype=float)
    hole1[-1] += 0.001
    hole2 = 7. * np.ones_like(sink_grid5.lake2, dtype=float)

    assert_array_almost_equal(
        sink_grid5.at_node["topographic__elevation"][sink_grid5.lake1], hole1
    )
    assert_array_almost_equal(
        sink_grid5.at_node["topographic__elevation"][sink_grid5.lake2], hole2
    )



#! /usr/bin/env python
"""Print user messages formatted landlab-style.

This module provides functions for printing nicely-formatted
messages to the user.  Messages are formatted in a particular
style so that all of landlab messages will have a similar
look. Anytime landlab prints something for an end-user to see,
this module should be used.

This module also provides convenience functions for print
particular types of messages. Warning and error messages,
for instance.

Examples
--------
>>> from __future__ import print_function

Oftentimes when writing code we may need to print a lengthy
message for the user. This may result in code that looks like
the following.

>>> message = ('Lorem ipsum dolor sit amet, consectetur '
...            'adipiscing elit, sed do eiusmod tempor '
...            'incididunt ut labore et dolore magna aliqua. '
...            'Ut enim ad minim veniam, quis nostrud exercitation '
...            'ullamco laboris nisi ut aliquip ex ea commodo '
...            'consequat.')

Printing this message string would result in one long line that
would, most likely, extend beyond the user's terminal and be
difficult to read. One solution would be to join the lines
by line separators but then that would result in a bunch of really
short lines.

To help with this, landlab provides a set of functions with the
most basic being `format_message`.

>>> from landlab.core.messages import format_message
>>> print(format_message(message))
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do
eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad
minim veniam, quis nostrud exercitation ullamco laboris nisi ut
aliquip ex ea commodo consequat.

landlab also provides functions for printing warning and error
messages.


>>> from landlab.core.messages import warning_message
>>> message = ('Lorem ipsum dolor sit amet, consectetur\\n'
...            'adipiscing elit, sed do eiusmod tempor\\n'
...            'incididunt ut labore et dolore magna aliqua.')
>>> print(warning_message(message))
WARNING
=======
<BLANKLINE>
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do
eiusmod tempor incididunt ut labore et dolore magna aliqua.

>>> from landlab.core.messages import error_message
>>> print(error_message(message))
ERROR
=====
<BLANKLINE>
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do
eiusmod tempor incididunt ut labore et dolore magna aliqua.

Another common design pattern used in landlab is to allow the
user, usually through a keyword, to control what happens
if a particular assertion fails. For instance, the user
may want the code to raise an error, or print a warning
message, or do nothing at all. The `assert_or_print` function
should be used in these cases.

>>> from landlab.core.messages import assert_or_print

>>> dt = 1e6
>>> assert_or_print(dt < 1, 'Unstable time step!', onerror='pass')
>>> assert_or_print(dt < 1, 'Unstable time step!', onerror='warn')
...     #doctest: +SKIP
WARNING
=======

Unstable time step!

>>> assert_or_print(dt < 1, 'Unstable time step!', onerror='raise')
...     #doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
...
AssertionError
"""
from __future__ import print_function

import os
import sys
import textwrap
import re

import six


def indent_and_wrap(content, indent=''):
    """Indent and wrap some text

    Lines are first dedented to remove common leading whitespace,
    then indented according to the value of *indent*, and then
    wrapped at 70 characters (indenting if necessary with subsequent
    indent being twice *indent*).

    Note that when looking for common whitespace, the first line is
    ignored.

    Parameters
    ----------
    content : str
        The content to wrap.

    Returns
    -------
    str
        The content properly wrapped and indented.

    Examples
    --------
    >>> from __future__ import print_function
    >>> from landlab.core.messages import indent_and_wrap
    >>> content = '''@book{knuth1998art,
    ...     title={The art of computer programming: sorting and searching},
    ...     author={Knuth, Donald Ervin},
    ...     volume={3},
    ...     year={1998},
    ...     publisher={Pearson Education}
    ...     }'''
    >>> print(indent_and_wrap(content))
    @book{knuth1998art,
    title={The art of computer programming: sorting and searching},
    author={Knuth, Donald Ervin},
    volume={3},
    year={1998},
    publisher={Pearson Education}
    }
    """
    wrapper = textwrap.TextWrapper(initial_indent=indent,
                                   subsequent_indent=2 * indent)
    lines = content.splitlines()
    first_line, the_rest = [lines[0].strip()], lines[1:]
    if the_rest:
        the_rest = textwrap.dedent(os.linesep.join(the_rest)).splitlines()
    if first_line[0]:
        lines = first_line + the_rest
    else:
        lines = the_rest
    return os.linesep.join([os.linesep.join(wrapper.wrap(line)) for line in lines])


def split_paragraphs(msg, linesep=os.linesep):
    """Split text into paragraphs.

    Split a block of text into paragraphs. A paragraph is
    defined as adjacent new-line characters (possibly separated
    by some whitespace).

    Parameters
    ----------
    msg : str
        Text to split into paragraphs.
    linesep : str, optional
        Line separator used in the message string.
    
    Returns
    -------
    list of str
        List of paragraphs.

    Examples
    --------
    >>> from landlab.core.messages import split_paragraphs
    >>> text = '''
    ... Pharetra pharetra massa massa ultricies mi quis hendrerit.
    ...
    ... Dictumst vestibulum rhoncus est pellentesque.
    ... '''
    >>> split_paragraphs(text, linesep='\\n') #doctest: +NORMALIZE_WHITESPACE
    ['Pharetra pharetra massa massa ultricies mi quis hendrerit.',
     'Dictumst vestibulum rhoncus est pellentesque.']

    >>> text = '''
    ... Pharetra pharetra massa massa ultricies mi quis hendrerit.
    ... Dictumst vestibulum rhoncus est pellentesque.
    ... '''
    >>> len(split_paragraphs(text, linesep='\\n'))
    1
    """
    pattern = linesep + '\s*' + linesep
    parsep = linesep * 2
    return re.sub(pattern, parsep, msg.strip()).split(parsep)


def format_message(msg, header=None, footer=None, linesep=os.linesep):
    """Format a message, landlab-style.

    Create a nicely formatted message that splits paragraphs,
    dedents paragraphs, and wraps text at 70 characters.
    Optionally, add a header and footer to the resulting message.


    Parameters
    ----------
    msg : str
        The message to be formatted.
    header : str or list of str, optional
        String to add before *msg*.
    footer : str or list of str, optional
        String to add after *msg*.

    Returns
    -------
    str
        The formatted message.

    Examples
    --------
    >>> from __future__ import print_function
    >>> from landlab.core.messages import format_message
    >>> text = '''
    ... Lorem ipsum dolor sit amet, consectetur
    ... adipiscing elit, sed do eiusmod tempor
    ... incididunt ut labore et dolore magna aliqua.
    ...
    ... Pharetra pharetra massa massa ultricies mi
    ... quis hendrerit.
    ...
    ... Dictumst vestibulum rhoncus est pellentesque.
    ... Sed viverra tellus in hac habitasse platea
    ... dictumst vestibulum rhoncus.'''
    >>> print(format_message(text))
    Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do
    eiusmod tempor incididunt ut labore et dolore magna aliqua.
    <BLANKLINE>
    Pharetra pharetra massa massa ultricies mi quis hendrerit.
    <BLANKLINE>
    Dictumst vestibulum rhoncus est pellentesque. Sed viverra tellus in
    hac habitasse platea dictumst vestibulum rhoncus.
    """
    if isinstance(header, six.string_types):
        header = [header]
    header = header or []

    if isinstance(footer, six.string_types):
        footer = [footer]
    footer = footer or []

    paragraphs = header
    if msg is not None:
        for paragraph in split_paragraphs(msg.strip(), linesep=linesep):
            paragraphs.append(
                os.linesep.join(textwrap.wrap(textwrap.dedent(paragraph))))
    paragraphs += footer

    return (os.linesep * 2).join(paragraphs)


def warning_message(msg=None, **kwds):
    """Create a warning message, landlab-style.

    Parameters
    ----------
    msg : str, optional
        Warning message.

    Returns
    -------
    str
        The formatted warning message.

    Examples
    --------
    >>> from __future__ import print_function
    >>> from landlab.core.messages import warning_message
    >>> print(warning_message('Dictumst vestibulum rhoncus est pellentesque.'))
    WARNING
    =======
    <BLANKLINE>
    Dictumst vestibulum rhoncus est pellentesque.
    """
    return format_message(msg,
                          header=os.linesep.join(['WARNING', '=======']),
                          **kwds)


def error_message(msg=None, **kwds):
    """Create an error  message, landlab-style.

    Parameters
    ----------
    msg : str, optional
        Warning message.

    Returns
    -------
    str
        The formatted warning message.

    Examples
    --------
    >>> from __future__ import print_function
    >>> from landlab.core.messages import error_message
    >>> print(error_message('Dictumst vestibulum rhoncus est pellentesque.'))
    ERROR
    =====
    <BLANKLINE>
    Dictumst vestibulum rhoncus est pellentesque.
    """
    return format_message(msg,
                          header=os.linesep.join(['ERROR', '=====']),
                          **kwds)


def assert_or_print(cond, msg=None, onerror='raise', file=sys.stdout):
    """Make an assertion printing a message if it fails.

    Specify an action to take if an assertion fails, depending on
    the values of *onerror*. *onerror* must be one of:
        *  "pass": do nothing if the assertion passes or fails.
        *  "warn": print a warning message if the assertion fails.
        *  "error": print an error message and raise an `AssertionError`
           on failure.

    Parameters
    ----------
    cond : expression
        An expression to test.
    msg : str, optional
        Message to print if the condition is not met.
    onerror: {'pass', 'warn', 'raise'}, optional
        What to do if the condition evaluates to `False`.
    file : file_like, optional
        File-like object where the message is printed.

    Examples
    --------
    >>> from landlab.core.messages import assert_or_print

    >>> assert_or_print(True, 'Lorem ipsum', onerror='pass')
    >>> assert_or_print(False, 'Lorem ipsum', onerror='pass')

    >>> assert_or_print(True, 'Lorem ipsum', onerror='warn')
    >>> assert_or_print(False, 'Lorem ipsum', onerror='warn')

    >>> assert_or_print(True, 'Lorem ipsum', onerror='raise')
    >>> assert_or_print(False, 'Lorem ipsum', onerror='raise')
    ...     #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    AssertionError
    """
    if onerror not in ('pass', 'warn', 'raise'):
        raise ValueError("onerror must be one of 'pass', 'warn', or 'raise'")

    try:
        assert(cond)
    except AssertionError:
        if onerror == 'warn':
            print(warning_message(msg), file=file, end='')
        elif onerror == 'raise':
            print(error_message(msg), file=file, end='')
            raise



"""
Unit tests for \
    landlab.components.soil_moisture.soil_moisture_dynamics
"""
import pytest
from numpy.testing import assert_array_almost_equal
import numpy as np


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e0, 10e0), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(sm):
    assert sm.name == 'Soil Moisture'


def test_input_var_names(sm):
    assert sorted(sm.input_var_names) == [
        'rainfall__daily_depth',
        'soil_moisture__initial_saturation_fraction',
        'surface__potential_evapotranspiration_rate',
        'vegetation__cover_fraction',
        'vegetation__live_leaf_area_index',
        'vegetation__plant_functional_type'
    ]


def test_output_var_names(sm):
    assert sorted(sm.output_var_names) == [
        'soil_moisture__root_zone_leakage',
        'soil_moisture__saturation_fraction',
        'surface__evapotranspiration',
        'surface__runoff',
        'vegetation__water_stress',
    ]


def test_var_units(sm):
    assert set(sm.input_var_names) | set(sm.output_var_names), set(dict(sm.units).keys())

    assert sm.var_units('vegetation__cover_fraction') == 'None'
    assert sm.var_units('vegetation__live_leaf_area_index') == 'None'
    assert sm.var_units('surface__potential_evapotranspiration_rate') == 'mm'
    assert sm.var_units('vegetation__plant_functional_type') == 'None'
    assert sm.var_units('vegetation__water_stress') == 'None'
    assert sm.var_units('soil_moisture__saturation_fraction') == 'None'
    assert sm.var_units('soil_moisture__initial_saturation_fraction') == 'None'
    assert sm.var_units('soil_moisture__root_zone_leakage') == 'mm'
    assert sm.var_units('surface__runoff') == 'mm'
    assert sm.var_units('surface__evapotranspiration') == 'mm'
    assert sm.var_units('rainfall__daily_depth') == 'mm'


def test_grid_shape(sm):
    assert sm.grid.number_of_node_rows == _SHAPE[0]
    assert sm.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(sm):
    assert sm.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(sm):
    assert sm.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(sm):
    for name in sm.grid['node']:
        field = sm.grid['node'][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (sm.grid.number_of_node_rows * sm.grid.number_of_node_columns, )
                      
    for name in sm.grid['cell']:
        field = sm.grid['cell'][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (sm.grid.number_of_cell_rows *
                               sm.grid.number_of_cell_columns, )

    with pytest.raises(KeyError):
        sm.grid['not_a_var_name']


def test_field_initialized_to_zero(sm):
    for name in sm.grid['node']:
        field = sm.grid['node'][name]
        assert_array_almost_equal(field, np.zeros(sm.grid.number_of_nodes))
    for name in sm.grid['cell']:
        field = sm.grid['cell'][name]
        assert_array_almost_equal(field, np.zeros(sm.grid.number_of_cells))



"""Test HexGraph and DualHexGraph."""
from pytest import approx
from numpy.testing import assert_array_equal, assert_array_almost_equal
import numpy as np

from landlab.graph import DualHexGraph


ROOT_3_OVER_2 = np.sqrt(3.) * .5


def test_create():
    """Test creating a dual hex graph with rectangular layout."""
    graph = DualHexGraph((4, 3), node_layout='rect')

    assert graph.number_of_nodes == 12
    assert graph.number_of_links == 23
    assert graph.number_of_patches == 12

    assert graph.number_of_corners == 10
    assert graph.number_of_faces == 11
    assert graph.number_of_cells == 2


def test_create_hex():
    """Test creating a dual hex graph with hex layout."""
    graph = DualHexGraph((4, 3), node_layout='hex')

    assert graph.number_of_nodes == 16
    assert graph.number_of_links == 34
    assert graph.number_of_patches == 19

    assert graph.number_of_corners == 19
    assert graph.number_of_faces == 23
    assert graph.number_of_cells == 5


def test_create_rect1():
    """Test creating a dual hex graph."""
    graph = DualHexGraph((4, 3), node_layout='rect1')

    assert graph.number_of_nodes == 14
    assert graph.number_of_links == 28
    assert graph.number_of_patches == 15

    assert graph.number_of_corners == 13
    assert graph.number_of_faces == 15
    assert graph.number_of_cells == 3


def test_spacing():
    """Test spacing of nodes."""
    graph = DualHexGraph((20, 31), node_layout='rect1')
    assert_array_almost_equal(graph.length_of_link, 1.)

    graph = DualHexGraph((31, 20), spacing=2, node_layout='rect1')
    assert_array_almost_equal(graph.length_of_link, 2.)


def test_origin():
    """Test setting the origin."""
    graph = DualHexGraph((4, 3))

    assert graph.y_of_node[0] == approx(0.)
    assert graph.x_of_node[0] == approx(0.)
    assert graph.x_of_corner[0] == approx(1.5)

    graph = DualHexGraph((4, 3), origin=(.5, .25))

    assert graph.y_of_node[0] == approx(.5)
    assert graph.x_of_node[0] == approx(.25)
    assert graph.x_of_corner[0] == approx(1.75)


def test_orientation():
    """Test vertical and horizontal orientation."""
    graph = DualHexGraph((3, 4), orientation='vertical')
    assert_array_almost_equal(graph.y_of_corner,
                              [.5, .5, 1, 1, 1, 1.5, 1.5, 1.5, 2, 2])

    graph = DualHexGraph((4, 3), orientation='horizontal')
    assert_array_almost_equal(graph.x_of_corner,
                              [1.5, 1, 2, 1, 2, .5, 1.5, .5, 1.5, 1])


def test_adjacent_corners_at_corner():
    graph = DualHexGraph((3, 3), node_layout='hex')
    assert_array_equal(graph.adjacent_corners_at_corner,
                       [[ 3,  2, -1],
                        [ 4 , 3, -1],
                        [ 5,  0, -1],
                        [ 6,  0,  1],
                        [ 7,  1, -1],
                        [ 8,  2, -1],
                        [ 9,  8,  3],
                        [ 9,  4, -1],
                        [ 5,  6, -1],
                        [ 6,  7, -1]])



import numpy as np

from ...utils.jaggedarray import unravel


def links_at_patch(patches, sort=True, nodes_at_link=None):
    """Construct as links_at_patch array for a graph.

    Parameters
    ----------
    patches : tuple of ndarray of int
        Links that define patches as `(links, offset_to_patch)`.
    sort : bool, optional
        Sort the links.
    nodes_at_link : ndarray of int, shape `(n_links, 2)`
        Nodes at link tail and head.

    Examples
    --------

    """
    from ..sort.ext.remap_element import reorder_links_at_patch
    from ..quantity.ext.of_link import calc_midpoint_of_link

    links_at_patch, offset_to_patch = patches

    sort = False
    # print 'SORTING TURNED OFF'
    if sort:
        xy_of_link = np.empty((len(nodes_at_link), 2), dtype=float)
        calc_midpoint_of_link(nodes_at_link, x_of_node, y_of_node, xy_of_link)
        reorder_links_at_patch(links_at_patch, offset_to_patch, xy_of_link)

        sort_patches(links_at_patch, offset_to_patch, xy_of_link)

    return unravel(links_at_patch, offset_to_patch, pad=-1)



#! /usr/bin/env python

import inspect

from .component import load_landlab_components
from landlab import Palette, Arena


class Error(Exception):
    """
    Base exception for this module
    """


class Framework(object):
    """
    A framework for connecting and running component from The Landlab.
    """

    def __init__(self):
        self._palette = Palette(**load_landlab_components())
        self._arena = Arena()

    def instantiate(self, name):
        """
        Instantiate a component called *name* from the palette and move it to
        the arena.
        """
        try:
            self._arena.instantiate(self._palette.get(name), name)
        except KeyError:
            pass

    def remove(self, name):
        """
        Remove a component called *name* from the arena.
        """
        try:
            self._arena.remove(name)
        except KeyError:
            pass

    def list_palette(self):
        """
        Get a list of names of the components in the palette.
        """
        return self._palette.list()

    def list_arena(self):
        """
        Get a list of names of the components in the arena.
        """
        return self._arena.list()

    def arena_uses(self):
        """
        Get a list of variable names that components in the arena use.
        """
        return self._arena.uses()

    def arena_provides(self):
        """
        Get a list of variable names that components in the arena provide.
        """
        return self._arena.provides()

    def palette_uses(self):
        """
        Get a list of variable names that components in the palette use.
        """
        return self._palette.uses()

    def palette_provides(self):
        """
        Get a list of variable names that components in the palette provide.
        """
        return self._palette.provides()

    def __repr__(self):
        return 'Framework(%s)' % ', '.join(self._palette.keys())



import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np

import landlab


def test_imshow_grid():
    rmg = landlab.RasterModelGrid(4, 5)

    pp = PdfPages('test.pdf')

    values = np.arange(rmg.number_of_nodes)
    landlab.plot.imshow_grid(rmg, values, values_at='node', limits=(0, 20))
    pp.savefig()

    plt.clf()
    rmg.status_at_node[7] = landlab.CLOSED_BOUNDARY
    values = np.arange(rmg.number_of_cells)
    landlab.plot.imshow_grid(rmg, values, values_at='cell',
                             symmetric_cbar=True)
    pp.savefig()
    pp.close()



"""Test StructuredQuadGraph."""
from numpy.testing import assert_array_equal, assert_array_almost_equal
import numpy as np

from landlab.graph import DualStructuredQuadGraph, DualUniformRectilinearGraph


def test_create():
    """Test creating a quad graph."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))

    assert graph.number_of_nodes == 9
    assert graph.number_of_links == 12
    assert graph.number_of_patches == 4

    assert graph.number_of_corners == 4
    assert graph.number_of_faces == 4
    assert graph.number_of_cells == 1


def test_create_raster():
    """Test creating a quad graph."""
    graph = DualUniformRectilinearGraph((3, 4), spacing=(2., 3.))

    assert graph.number_of_nodes == 12
    assert graph.number_of_links == 17
    assert graph.number_of_patches == 6

    assert graph.number_of_corners == 6
    assert graph.number_of_faces == 7
    assert graph.number_of_cells == 2


def test_raster_spacing():
    """Test the spacing keyword for raster."""
    graph = DualUniformRectilinearGraph((3, 4), spacing=(2., 3.))

    assert_array_equal(graph.length_of_link,
                       [3., 3., 3., 2., 2., 2., 2.,
                        3., 3., 3., 2., 2., 2., 2.,
                        3., 3., 3.])
    assert_array_equal(graph.length_of_face,
                       [3., 3., 2., 2., 2., 3., 3.])


def test_raster_spacing_as_scalar():
    """Test the spacing keyword as a scalar for raster."""
    graph = DualUniformRectilinearGraph((3, 4), spacing=2.)

    assert_array_equal(graph.length_of_link,
                       [2., 2., 2., 2., 2., 2., 2.,
                        2., 2., 2., 2., 2., 2., 2.,
                        2., 2., 2.])
    assert_array_equal(graph.length_of_face,
                       [2., 2., 2., 2., 2., 2., 2.])


def test_raster_origin():
    """Test the origin keyword for raster."""
    graph = DualUniformRectilinearGraph((3, 4), origin=(-1., 10.))

    assert_array_equal(graph.xy_of_node[:, 0],
                       [10., 11., 12., 13.,
                        10., 11., 12., 13.,
                        10., 11., 12., 13.])
    assert_array_equal(graph.xy_of_node[:, 1],
                       [-1., -1., -1., -1.,
                         0.,  0.,  0.,  0.,
                         1.,  1.,  1.,  1.])

    assert_array_equal(graph.xy_of_corner[:, 0],
                       [10.5, 11.5, 12.5,
                        10.5, 11.5, 12.5])
    assert_array_equal(graph.xy_of_corner[:, 1],
                       [-.5, -.5, -.5,
                         .5,  .5,  .5])


def test_raster_origin_as_scalar():
    """Test the origin keyword as a scalar for raster."""
    graph = DualUniformRectilinearGraph((3, 4), origin=-1.)

    assert_array_equal(graph.xy_of_node[:, 0],
                       [-1., 0., 1., 2.,
                        -1., 0., 1., 2.,
                        -1., 0., 1., 2.])
    assert_array_equal(graph.xy_of_node[:, 1],
                       [-1., -1., -1., -1.,
                         0.,  0.,  0.,  0.,
                         1.,  1.,  1.,  1.])

    assert_array_equal(graph.xy_of_corner[:, 0],
                       [-.5, .5, 1.5,
                        -.5, .5, 1.5])
    assert_array_equal(graph.xy_of_corner[:, 1],
                       [-.5, -.5, -.5,
                         .5,  .5,  .5])

def test_perimeter_corners():
    """Test the perimeter corners."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.perimeter_corners, [1, 3, 2, 0])


def test_length_of_face():
    """Test length of faces."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))

    assert_array_almost_equal(graph.length_of_face, [1.5, 1.5, 1.5, 1.5])
    assert_array_almost_equal(graph.width_of_face,[1.5, 1.5, 1.5, 1.5])
    assert graph.length_of_face is graph.width_of_face


def test_area_of_cell():
    """Test areas of patches."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_almost_equal(graph.area_of_cell, [2.25])


def test_corners_at_cell():
    """Test corners of cells."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.corners_at_cell, [[3, 2, 0, 1]])


def test_cells_at_corner():
    """Test areas of patches."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.cells_at_corner,
                       [[ 0, -1, -1, -1], [-1,  0, -1, -1],
                        [-1, -1, -1,  0], [-1, -1,  0, -1]])


def test_cells_at_face():
    """Test cells on either side of faces."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.cells_at_face,
                       [[-1,  0], [ 0, -1], [-1,  0], [ 0, -1]])


def test_faces_at_cell():
    """Test faces that form cells."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.faces_at_cell, [[2, 3, 1, 0]])


def test_corners_at_face():
    """Test corners at face tail and head."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.corners_at_face,
                       [[0, 1], [0, 2], [1, 3], [2, 3]])


def test_faces_at_corner():
    """Test faces around corners."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.faces_at_corner,
                       [[ 0,  1, -1, -1], [-1, 2,  0, -1],
                        [ 3, -1, -1,  1], [-1, -1,  3, 2]])


def test_face_dirs_at_corner():
    """Test face directions at corners."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.face_dirs_at_corner,
                       [[-1, -1,  0,  0], [ 0, -1,  1,  0],
                        [-1,  0,  0,  1], [ 0,  0,  1,  1]])


def test_cell_at_node():
    """Test cell-node connectivity."""
    y = [0, 1, 3, 0, 1, 3, 0, 1, 3]
    x = [3, 3, 3, 4, 4, 4, 6, 6, 6]
    graph = DualStructuredQuadGraph((y, x), shape=(3, 3))
    assert_array_equal(graph.cell_at_node, [-1, -1, -1, -1, 0, -1, -1, -1, -1])

    graph = DualUniformRectilinearGraph((3, 4))
    assert_array_equal(graph.cell_at_node,
                       [-1, -1, -1, -1,
                        -1,  0,  1, -1,
                        -1, -1, -1, -1])


def test_link_at_face():
    """Test link-face connectivity."""
    graph = DualUniformRectilinearGraph((3, 4))
    assert_array_equal(graph.link_at_face, [4, 5, 7, 8, 9, 11, 12])
    assert_array_equal(graph.face_at_link,
                       [-1, -1, -1,
                        -1,  0,  1, -1,
                         2,  3,  4,
                        -1,  5,  6, -1,
                        -1, -1, -1])


def test_corner_at_face():
    """Test corner-face connectivity."""
    graph = DualUniformRectilinearGraph((3, 4))
    assert_array_equal(graph.corners_at_face,
                       [[0, 1], [1, 2],
                        [0, 3], [1, 4], [2, 5],
                        [3, 4], [4, 5]])
    assert_array_equal(graph.corner_at_face_tail, [0, 1, 0, 1, 2, 3, 4])
    assert_array_equal(graph.corner_at_face_head, [1, 2, 3, 4, 5, 4, 5])



#! /usr/bin/env python
import os

import pytest
import numpy as np
from numpy.testing import assert_array_almost_equal

from landlab.io import write_esri_ascii, read_esri_ascii
from landlab import RasterModelGrid


def test_grid_with_no_fields(tmpdir):
    grid = RasterModelGrid((4, 5), spacing=(2., 2.))
    with tmpdir.as_cwd():
        with pytest.raises(ValueError):
            write_esri_ascii('test.asc', grid)


def test_grid_with_one_field(tmpdir):
    grid = RasterModelGrid((4, 5), spacing=(2., 2.))
    grid.add_field('node', 'air__temperature', np.arange(20.))
    with tmpdir.as_cwd():
        files = write_esri_ascii('test.asc', grid)
        assert files == ['test.asc']
        for fname in files:
            assert os.path.isfile(fname)


def test_grid_with_two_fields(tmpdir):
    grid = RasterModelGrid((4, 5), spacing=(2., 2.))
    grid.add_field('node', 'air__temperature', np.arange(20.))
    grid.add_field('node', 'land_surface__elevation', np.arange(20.))
    with tmpdir.as_cwd():
        files = write_esri_ascii('test.asc', grid)
        files.sort()
        assert files == ['test_air__temperature.asc',
                         'test_land_surface__elevation.asc']
        for fname in files:
            assert os.path.isfile(fname)


def test_names_keyword_as_str(tmpdir):
    grid = RasterModelGrid((4, 5), spacing=(2., 2.))
    grid.add_field('air__temperature', np.arange(20.), at='node')
    grid.add_field('land_surface__elevation', np.arange(20.), at='node')

    with tmpdir.as_cwd():
        files = write_esri_ascii('test.asc', grid, names='air__temperature')
        assert files == ['test.asc']
        assert os.path.isfile('test.asc')


def test_names_keyword_as_list(tmpdir):
    grid = RasterModelGrid((4, 5), spacing=(2., 2.))
    grid.add_field('air__temperature', np.arange(20.), at='node')
    grid.add_field('land_surface__elevation', np.arange(20.), at='node')

    with tmpdir.as_cwd():
        files = write_esri_ascii('test.asc', grid, names=['air__temperature'])
        assert files == ['test.asc']
        assert os.path.isfile('test.asc')


def test_names_keyword_multiple_names(tmpdir):
    grid = RasterModelGrid((4, 5), spacing=(2., 2.))
    grid.add_field('node', 'air__temperature', np.arange(20.))
    grid.add_field('node', 'land_surface__elevation', np.arange(20.))

    with tmpdir.as_cwd():
        files = write_esri_ascii('test.asc', grid,
                                 names=['air__temperature',
                                        'land_surface__elevation'])
        files.sort()
        assert files == ['test_air__temperature.asc',
                         'test_land_surface__elevation.asc']
        for fname in files:
            assert os.path.isfile(fname)


def test_names_keyword_with_bad_name(tmpdir):
    grid = RasterModelGrid((4, 5), spacing=(2., 2.))
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with tmpdir.as_cwd():
        with pytest.raises(ValueError):
            write_esri_ascii('test.asc', grid, names='not_a_name')


def test_clobber_keyword(tmpdir):
    grid = RasterModelGrid((4, 5), spacing=(2., 2.))
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with tmpdir.as_cwd():
        write_esri_ascii('test.asc', grid)
        with pytest.raises(ValueError):
            write_esri_ascii('test.asc', grid)
        with pytest.raises(ValueError):
            write_esri_ascii('test.asc', grid, clobber=False)
        write_esri_ascii('test.asc', grid, clobber=True)


def test_write_then_read(tmpdir):
    grid = RasterModelGrid((4, 5), spacing=(2., 2.))
    grid.add_field('node', 'air__temperature', np.arange(20.))

    with tmpdir.as_cwd():
        write_esri_ascii('test.asc', grid)
        new_grid, field = read_esri_ascii('test.asc')

    assert grid.number_of_node_columns == new_grid.number_of_node_columns
    assert grid.number_of_node_rows == new_grid.number_of_node_rows
    assert grid.dx == new_grid.dx
    assert_array_almost_equal(grid.node_x, new_grid.node_x)
    assert_array_almost_equal(grid.node_y, new_grid.node_y)
    assert_array_almost_equal(field, grid.at_node['air__temperature'])



"""
Source Tracking Algorithm
+++++++++++++++++++++++++
.. autosummary::
    :toctree: generated/
    ~landlab.utils.source_tracking_algorithm.convert_arc_flow_directions_to_landlab_node_ids
    ~landlab.utils.source_tracking_algorithm.track_source
    ~landlab.utils.source_tracking_algorithm.find_unique_upstream_hsd_ids_and_fractions
"""
"""
Authors: Sai Nudurupati & Erkan Istanbulluoglu

Ref 1: 'The Landlab LandslideProbability Component User Manual' @
https://github.com/RondaStrauch/pub_strauch_etal_esurf/blob/master/LandslideComponentUsersManual.docx

MD - Modeling Domain - Raster grid that is being analyzed/worked upon.
HSD - Hydrologic Source Domain - Grid that is at least as coarse as MD. For
      more info, refer Ref 1
"""

# %%
# Import required libraries
import numpy as np
import copy
from collections import Counter

# %%
def convert_arc_flow_directions_to_landlab_node_ids(grid, flow_dir_arc):
    """Convert Arc flow_directions to RasterModelGrid node ids

    This function receives flow directions (D8) from ESRI ArcGIS and converts
    them to Landlab's RasterModelGrid node id. ESRI ArcGIS D8 flow directions
    are either of the eight valid output directions relating to the eight
    adjacent cells into which flow could travel. The valid output directions
    are powers of 2 starting from 2^0 (1) in the Eastern neighbor going
    clockwise to 2^7 (128) at Northeastern neighbor. For more information
    refer 'http://pro.arcgis.com/en/pro-app/tool-reference/spatial-analyst/
    how-flow-direction-works.htm'
    
    Parameters
    ----------
    grid: RasterModelGrid
        A grid.
    flow_dir_arc: ndarray of int, shape (n_nodes, )
        flow directions derived from ESRII ArcGIS.
    
    Returns
    -------
    receiver_nodes: ndarray of int, shape (n_nodes, )
        downstream node at each node. Note that this array gives the
        receiver nodes only for the core nodes. For non-core
        nodes, a zero is used.  
    """
    r_arc_raw = np.log2(flow_dir_arc)
    r_arc_raw = r_arc_raw.astype('int')
    neigh_ = grid.neighbors_at_node
    diag_ = grid.diagonals_at_node
    neigh_ = np.fliplr(neigh_)
    diag_ = np.fliplr(diag_)
    a_n = np.hsplit(neigh_, 4)
    a_d = np.hsplit(diag_, 4)
    neighbors = np.hstack((a_n[-1], a_d[0], a_n[0], a_d[1], a_n[1], a_d[2],
                           a_n[2], a_d[3]))
    # Now neighbors has node ids of neighboring nodes in cw order starting at
    # right, hence the order of neighbors = [r, br, b, bl, l, tl, t, tr]
    receiver_nodes = np.zeros(grid.number_of_nodes, dtype=int)
    receiver_nodes[grid.core_nodes] = np.choose(r_arc_raw[grid.core_nodes],
                                   np.transpose(neighbors[grid.core_nodes]))
    return (receiver_nodes)


# %%
# Source Routing Algorithm
# Note 1: This algorithm works on core nodes only because core nodes
# have neighbors that are real values and not -1s.
# Note 2: Nodes in the following comments in this section refer to core nodes.
def track_source(grid, hsd_ids, flow_directions=None):
    """Track all contributing upstream core nodes for each core node

    This algorithm traverses the grid based on information of flow directions
    at nodes and at every node identifies all the nodes upstream of a given
    node. The algorithm creates a dictionary with an entry for each node;
    a node's entry in the dictionary will contain a list with the node_ids
    of all upstream nodes. Thus this method permits identification of the
    source area contributing to each and every node in the model grid. This
    function is different from a standard flow accumulation routine in that
    it not only calculates the amount of flow at each node, but records the
    IDs of all upstream nodes. However, similar to a standard
    flow accumulation routine, it produces an at_node array of the amount
    of flow passing through the node. It also differs from a standard
    flow accumulation routing in that it permits the mapping of flow inputs
    from a coarser grid to to a finer model grid.

    In its present implementation, the algorithm has not been optimized
    for efficient time use. Its methods are brute force and it should be
    expected to be time intensive. It is not recommended to be run frequently
    in a modeling exercise. Due to its intensive nature, this algorithm may
    fail with large watersheds (a present, the development team has not
    derived a maximum stable watershed size).

    This function was initially developed to find contributing area of a
    30 m grid (MD), where the quantitative data that we were interested in was
    available in significantly coarser resolution (called Hydrologic Source
    Domain (HSD)). Therefore, we started working with re-sampled HSD,
    that is at the same resolution as MD, and represents exactly the same
    landscape. Alternatively, one can use the node ids of MD
    (grid.nodes.flatten()) as input for hsd_ids.
    
    For more information, refer Ref 1.
    
    Parameters
    ----------
    grid: RasterModelGrid
        A grid.
    hsd_ids: ndarray of int, shape (n_nodes, )
        array that maps the nodes of the grid to, possibly coarser,
        Hydrologic Source Domain (HSD) grid ids.
    flow_directions: ndarray of int, shape (n_nodes, ), optional.
        downstream node at each node. Alternatively, this data can be
        provided as a nodal field 'flow__receiver_node' on the grid.
    
    Returns
    -------
    (hsd_upstr, flow_accum): (dictionary, ndarray of shape (n_nodes))
        'hsd_upstr' maps each grid node to corresponding
        contributing upstream hsd_ids. hsd_upstr.keys() will return
        node_ids of the grid. hsd_upstr.values() will return lists of
        all upstream contributing hsd_ids, including repitions of hsd_ids,
        at corresponding node_ids.
        'flow_accum' is an array of the number of upstream contributing
        nodes at each node.
    """   
    if flow_directions is None:
        r = grid.at_node['flow__receiver_node']
    else:
        r = flow_directions
    z = grid.at_node['topographic__elevation']
    core_nodes = grid.core_nodes
    core_elev = z[core_nodes]
    # Sort all nodes in the descending order of elevation
    sor_z = core_nodes[np.argsort(core_elev)[::-1]]
    # Create a list to record all nodes that have been visited
    # To store nodes that have already been counted
    alr_counted = []
    flow_accum = np.zeros(grid.number_of_nodes, dtype=int)
    hsd_upstr = {}
    # Loop through all nodes
    for i in sor_z:
        # Check 1: Check if this node has been visited earlier. If yes,
        # then skip to next node
        if i in alr_counted:
            continue
        # Check 2: If the visited node is a sink
        if r[i] == i:
            hsd_upstr.update({i: [hsd_ids[i]]})
            flow_accum[i] += 1.
            alr_counted.append(i)
            continue
        # Check 3: Now, if the node is not a sink and hasn't been visited, it
        # belongs to a stream segment. Hence, all the nodes in the stream will
        # have to betraversed.
        # stream_buffer is a list that will hold the upstream contributing
        # node information for that particular segment until reaching outlet.
        stream_buffer = []
        j = i
        switch_i = True
        a = 0.
        # Loop below will traverse the segment of the stream until an outlet
        # is reached.
        while True:
            # Following if loop is to execute the contents once the first node
            # in the segment is visited.
            if not switch_i:
                j = r[j]
                if j not in core_nodes:
                    break
            # If this node is being visited for the first time,
            # this 'if statement' will executed.
            if flow_accum[j] == 0.:
                a += 1.
                alr_counted.append(j)
                stream_buffer.append(hsd_ids[j])
            # Update number of upstream nodes.
            flow_accum[j] += a
            # If the node is being visited for the first time, the dictionary
            # 'hsd_upstr' will be updated.
            if j in hsd_upstr.keys():
                hsd_upstr[j] += copy.copy(stream_buffer)
            # If the node has been already visited, then the upstream segment
            # that was not accounted for in the main stem, would be added to
            # all downstream nodes, one by one, until the outlet is reached.
            else:
                hsd_upstr.update({j: copy.copy(stream_buffer)})
            # If the outlet is reached, the 'while' loop will be exited.
            if r[j] == j:
                break
            # This will be executed only for the first node of the
            # stream segment.
            if switch_i:
                switch_i = False
    return (hsd_upstr, flow_accum)


# %%
# Algorithm to calculate coefficients of each upstream HSD ID
def find_unique_upstream_hsd_ids_and_fractions(hsd_upstr):
    """Finds unique entries in hsd_upstr.values()

    This function operates on hsd_upstr.values(), that are lists of hsd_ids.
    Two new Python dictionaries, 'unique_ids' and 'fractions' are created.

    unique_ids.keys() = hsd_upstr.keys()
    unique_ids.values()[i] = list of unique entries in hsd_upstr.values()[i]

    fractions.keys() = hsd_upstr.keys()
    fractions.values()[i] = (number of entries of each unique_id.values()[i]/
    length of hsd_upstr.values()[i]) for each unique_id.values()[i] in the
    same order.

    Note that 'hsd_upstr' is the output of track_source(). You can use
    an alternative input. In that case, please refer to the documentation
    of track_source() or refer source_tracking_algorithm_user_manual for
    more information.
    
    Parameters
    ----------
    hsd_upstr: dictionary
        'hsd_upstr' maps each MD grid node to corresponding
        contributing upstream HSD ids.
        
    Returns
    -------
    (unique_ids, fractions): (dictionary, dictionary)
        Tuple of data. 'unique_ids' maps each MD node with all upstream HSD
        ids without repitition. 'fractions' maps each MD node with the
        fractions of contributions of the corresponding upstream HSD ids in
        the same order as uniques_ids[node_id].
    """
    unique_ids = {}  # Holds unique upstream HSD ids
    C = {}  # Holds corresponding total numbers
    fractions = {}  # Holds corresponding fractions of contribution
    for ke in hsd_upstr.keys():
        cnt = Counter()
        for num in hsd_upstr[ke]:
            cnt[num] += 1
        unique_ids.update({ke: cnt.keys()})
        buf = []
        for k in cnt.keys():
            buf.append(cnt[k])
        C.update({ke: buf})
        e = [s/float(sum(buf)) for s in buf]
        fractions.update({ke: e})
    return (unique_ids, fractions)



"""Generate precipitation using the Poisson pulse model.

Landlab component that generates precipitation events using the rectangular
Poisson pulse model described in Eagleson (1978, Water Resources Research).


No particular units must be used, but it was written with the storm units in
hours (hr) and depth units in millimeters (mm)

Written by Jordan Adams, 2013, updated May 2016
"""


import random
import numpy as np
from landlab import Component, ModelGrid
from six import next


class PrecipitationDistribution(Component):

    """Generate precipitation events.

    This component can generate a random storm duration, interstorm
    duration, precipitation intensity or storm depth from a Poisson
    distribution when given a mean value.

    Examples
    --------
    >>> from landlab.components.uniform_precip import PrecipitationDistribution
    >>> import numpy as np
    >>> np.random.seed(np.arange(10))

    To use hard-coded values for mean storm, mean interstorm, mean depth,
    model run time and delta t...  Say we use 1.5 for mean storm, 15 for mean
    interstorm, 0.5 for mean depth, 100 for model run time and 1 for delta t...

    >>> precip = PrecipitationDistribution(mean_storm_duration=1.5,
    ...     mean_interstorm_duration=15.0, mean_storm_depth=0.5,
    ...     total_t=100.0, delta_t=1.)
    >>> for (dt, rate) in precip.yield_storm_interstorm_duration_intensity():
    ...     pass  # and so on

    Alternatively, we can pass a grid to the component, and call yield_storms()
    to generate storm-interstorm float pairs while the intensity data is stored
    in the grid scalar field 'rainfall__flux':

    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((4, 5), (1., 1.))
    >>> precip = PrecipitationDistribution(mg, mean_storm_duration=1.5,
    ...     mean_interstorm_duration=15.0, mean_storm_depth=0.5,
    ...     total_t=46.)
    >>> storm_dts = []
    >>> interstorm_dts = []
    >>> intensities = []
    >>> precip.seed_generator(seedval=1)
    >>> for (storm_dt, interstorm_dt) in precip.yield_storms():
    ...     storm_dts.append(storm_dt)
    ...     interstorm_dts.append(interstorm_dt)
    ...     intensities.append(mg.at_grid['rainfall__flux'])
    >>> len(storm_dts) == 4  # 4 storms in the simulation
    True
    >>> len(interstorm_dts) == len(storm_dts)
    True
    >>> rf_intensities_to_test = np.array([0.8138257984406472,
    ...                                    0.15929112025199238,
    ...                                    0.17254519305000884,
    ...                                    0.09817611240558813])
    >>> np.allclose(intensities, rf_intensities_to_test)
    True
    >>> np.isclose(sum(storm_dts) + sum(interstorm_dts), 46.)  # test total_t
    True
    >>> np.isclose(interstorm_dts[-1], 0.)  # sequence truncated as necessary
    True

    We can also turn the generator straight into a list, like this:

    >>> precip.seed_generator(seedval=1)
    >>> steps = [(dt + istorm_dt) for (dt, istorm_dt) in precip.yield_storms()]
    >>> np.isclose(sum(steps), 46.)
    True
    """

    _name = 'PrecipitationDistribution'

    _input_var_names = tuple()

    _output_var_names = tuple()

    _optional_var_names = ('rainfall__flux', )

    _var_units = {'rainfall__flux': '[depth unit]/[time unit]', }

    _var_mapping = {'rainfall__flux': 'grid', }

    _var_doc = {
        'rainfall__flux':
            'Depth of water delivered per unit time in each storm', }

    def __init__(self, grid=None,
                 mean_storm_duration=0.0, mean_interstorm_duration=0.0,
                 mean_storm_depth=0.0, total_t=0.0, delta_t=None,
                 random_seed=0, **kwds):
        """Create the storm generator.

        Parameters
        ----------
        grid : ModelGrid
            A Landlab grid (optional). If provided, storm intensities will be
            stored as a grid scalar field as the component simulates storms.
        mean_storm_duration : float
            Average duration of a precipitation event.
        mean_interstorm_duration : float
            Average duration between precipitation events.
        mean_storm_depth : float
            Average depth of precipitation events.
        total_t : float, optional
            If generating a time series, the total amount of time.
        delta_t : float or None, optional
            If you want to break up storms into determined subsections using
            yield_storm_interstorm_duration_intensity, a delta_t is needed.
        random_seed : int or float, optional
            Seed value for random-number generator.
        """

        self.mean_storm_duration = mean_storm_duration

        self.mean_interstorm_duration = mean_interstorm_duration

        self.mean_storm_depth = mean_storm_depth

        self.run_time = total_t

        self.delta_t = delta_t

        if self.delta_t == 0.:
            self.delta_t = None

        # Mean_intensity is not set by the MPD, but can be drawn from
        # the mean storm depth and mean storm duration.
        self.mean_intensity = self.mean_storm_depth / self.mean_storm_duration

        # If a time series is created later, this blank list will be used.
        self.storm_time_series = []

        # Seed the random-number generator
        self.seed_generator(random_seed)

        # Given the mean values assigned above using either the model
        # parameter dictionary or the init function, we can call the
        # different methods to assign values from the Poisson distribution.

        self.storm_duration = self.get_precipitation_event_duration()
        self.interstorm_duration = self.get_interstorm_event_duration()
        self.storm_depth = self.get_storm_depth()
        self._elapsed_time = 0.

        # Test if we got a grid. If we did, then assign it to _grid, and we
        # are able to use the at_grid field. If not, that's cool too.
        if grid is not None:
            assert isinstance(grid, ModelGrid)  # must be a grid
            self._grid = grid

        # build LL fields, if a grid is supplied:
        if grid is not None:
            self.grid.add_field('grid', 'rainfall__flux', 0.)
            self._gridupdate = True
        else:
            self._gridupdate = False

        self._intensity = self.get_storm_intensity()

    def update(self):
        """Update the storm values.

        If new values for storm duration, interstorm duration, storm depth
        and intensity are needed, this method can be used to update those
        values one time.

        Examples
        --------
        >>> from landlab.components import PrecipitationDistribution
        >>> precip = PrecipitationDistribution(mean_storm_duration=1.5,
        ...     mean_interstorm_duration=15.0, mean_storm_depth=0.5,
        ...     total_t=100.0, delta_t=1)

        Additionally, if we wanted to update several times, a loop could be
        utilized to accomplish this. Say we want 5 storm_durations; this
        pseudo-code represents a way to accomplish this...

        >>> storm_duration_list = []
        >>> i = 0
        >>> while i < 4:
        ...     storm_duration_list.append(precip.storm_duration)
        ...     precip.update()
        ...     i += 1

        Note though that alternatively we could also do this, avoiding the
        method entirely...

        >>> from six import next
        >>> # ^^this lets you "manually" get the next item from the iterator
        >>> precip = PrecipitationDistribution(mean_storm_duration=1.5,
        ...     mean_interstorm_duration=15.0, mean_storm_depth=0.5,
        ...     total_t=46.)
        >>> storm_duration_list = []
        >>> for i in range(5):
        ...     storm_duration_list.append(next(
        ...         precip.yield_storm_interstorm_duration_intensity())[0])

        Notice that doing this will *not* automatically stop the iteration,
        however - it will continue ad infinitum.
        """
        self.storm_duration = self.get_precipitation_event_duration()
        self.interstorm_duration = self.get_interstorm_event_duration()
        self.storm_depth = self.get_storm_depth()
        self._intensity = self.get_storm_intensity()

    def get_precipitation_event_duration(self):
        """This method is the storm generator.

        This method has one argument: the mean_storm_duration parameter.
        (In Eagleson (1978), this parameter was called Tr.)

        It finds a random storm_duration value
        based on the poisson distribution about the mean.
        This is accomplished using the expovariate function
        from the "random" standard library.

        Returns
        -------
        float
            The storm duration.
        """
        return random.expovariate(1.0 / self.mean_storm_duration)

    def get_interstorm_event_duration(self):
        """Generate interstorm events.

        This method takes one argument, the mean_interstorm_duration parameter.
        (In Eagleson (1978), this parameter was called Tb.)

        This method is modeled identically to
        get_precipitation_event_duration()

        This method finds a random value for interstorm_duration
        based on the poisson distribution about the mean.
        This is accomplished using the expovariate function
        from the "random" standard library.

        Returns
        -------
        float
            The interstorm duration.
        """
        return random.expovariate(1.0 / self.mean_interstorm_duration)

    def get_storm_depth(self):
        """Generate storm depth.

        Storm depth is used to generate a realistic
        intensity for different storm events.

        (In Eagleson (1978) this parameter was called "h")

        This method requires storm_duration, mean_storm_duration
        and the mean_storm_depth. Storm_duration is generated through
        the initialize() or update() method.

        Numpy has a random number generator to get values
        from a given Gamma distribution. It takes two arguments,
        alpha (or the shape parameter), which is the generated over the mean
        event and beta (or the scale parameter), which is the mean value
        These are all arguments in the function, which returns storm depth.

        Returns
        -------
        float
            The storm depth.
        """

        shape_parameter = (self.storm_duration / self.mean_storm_duration)
        scale_parameter = (self.mean_storm_depth)
        self.storm_depth = np.random.gamma(shape_parameter, scale_parameter)
        return self.storm_depth

    def get_storm_intensity(self):
        """Get the storm intensity.

        This method draws storm intensity out of the storm depth generated by
        get_storm_depth.

        This method requires the storm_depth and storm_duration
        and is the same as the parameter ("i") in Eagleson (1978), but instead
        of being drawn from Poission, this is drawn from the Gamma distribution
        of (*h*), as :math:`h = i * Tr`.

        Returns
        -------
        float
            The storm intensity.
        """
        self._intensity = self.storm_depth / self.storm_duration
        if self._gridupdate:
            self.grid.at_grid['rainfall__flux'] = self._intensity
        return self._intensity

    def get_storm_time_series(self):
        """Get a time series of storms.

        This method creates a time series of storms based on storm_duration,
        and interstorm_duration. From these values it will calculate a complete
        time series.

        The storm_time_series returned by this method is made up of sublists,
        each comprising of three sub-parts (e.g. [[x,y,z], [a,b,c]]) where x
        and a are the beginning times of a precipitation event, y and b are the
        ending times of the precipitation event and z and c represent the
        average intensity (mm/hr) of the storm lasting from x to y and a to be,
        respectively.

        Even if a grid was passed to the component at instantiation, calling
        this method does not update the grid fields.

        Returns
        -------
        array
            containing several sub-arrays of events [start, finish, intensity]
        """

        storm = self.get_precipitation_event_duration()
        self.get_storm_depth()
        intensity = self.get_storm_intensity()
        self.storm_time_series.append([0, storm, intensity])

        storm_helper = storm
        storm_iterator = storm
        while storm_iterator <= self.run_time:
            next_storm_start = storm_helper + (round(
                                    self.get_interstorm_event_duration(), 2))
            next_storm_end = next_storm_start + (round(
                                self.get_precipitation_event_duration(), 2))
            intensity = round(self.get_storm_intensity(), 2)
            self.get_storm_depth()
            self.storm_time_series.append(
                            [next_storm_start, next_storm_end, intensity])
            storm_iterator = storm_helper
            storm_helper = next_storm_end
            storm_iterator = storm_helper
        return self.storm_time_series

    def yield_storm_interstorm_duration_intensity(self,
                                                  subdivide_interstorms=False):
        """Iterator for a time series of storms interspersed with interstorms.

        This method is intended to be equivalent to get_storm_time_series,
        but instead offers a generator functionality. This will be useful in
        cases where the whole sequence of storms and interstorms doesn't need
        to be stored, where we can save memory this way.

        The method keeps track of the delta_t such that if a storm needs to be
        generated longer than this supplied model timestep, the generator will
        return the storm in "chunks", until there is no more storm duration.
        e.g.,
        storm of intensity 1. is 4.5 long, the delta_t is 2., the generator
        yields (2.,1.) -> (2.,1.) -> (0.5,1.) -> ...

        If delta_t is None or not supplied, no subdivision occurs.

        Once a storm has been generated, this method will follow it with the
        next interstorm, yielded as (interstorm_duration, 0.). Note that the
        interstorm will NOT be subdivided according to delta_t unless you set
        the flag *subdivide_interstorms* to True.

        The method will keep yielding until it reaches the RUN_TIME, where it
        will terminate.

        Yields
        ------
        tuple of float
            (interval_duration, rainfall_rate_in_interval)

        Notes
        -----
        One recommended procedure is to instantiate the generator, then call
        instance.next() (in Python 2) or next(instance) (in Python 3)
        repeatedly to get the sequence.
        """
        # Added DEJH, Dec 2014
        # Modified to use an optional output field, DEJH 1/8/17

        delta_t = self.delta_t
        if delta_t is None:
            assert subdivide_interstorms is False, (
                'You specified you wanted storm subdivision, but did not ' +
                'provide a delta_t to allow this!')
        self._elapsed_time = 0.
        while self._elapsed_time < self.run_time:
            storm_duration = self.get_precipitation_event_duration()
            step_time = 0.
            self.get_storm_depth()
            self._intensity = self.get_storm_intensity()  # this is a rate
            # ^ this updates the grid field, if needed
            if self._elapsed_time + storm_duration > self.run_time:
                storm_duration = self.run_time - self._elapsed_time
            while delta_t is not None and storm_duration - step_time > delta_t:
                yield (delta_t, self._intensity)
                step_time += delta_t
            yield (storm_duration - step_time, self._intensity)
            self._elapsed_time += storm_duration

            # If the last storm did not use up all our elapsed time, generate
            # an inter-storm period.
            if self._elapsed_time < self.run_time:
                interstorm_duration = self.get_interstorm_event_duration()
                if self._elapsed_time + interstorm_duration > self.run_time:
                    interstorm_duration = self.run_time - self._elapsed_time
                self._intensity = 0.
                if self._gridupdate:
                    self.grid.at_grid['rainfall__flux'] = 0.
                if subdivide_interstorms:
                    step_time = 0.
                    while interstorm_duration-step_time > delta_t:
                        yield (delta_t, 0.)
                        step_time += delta_t
                    yield (interstorm_duration - step_time, 0.)
                else:
                    yield (interstorm_duration, 0.)
                self._elapsed_time += interstorm_duration

    def yield_storms(self):
        """Iterator for a time series of storm-interstorm pairs.

        This method is very similar to this component's other generator,
        yield_storm_interstorm_duration_intensity(), but the way it yields is
        slightly different. Instead of yielding (interval_duration, rf_rate),
        with interstorms represented as intervals with rf_rate = 0, it yields:

            (storm_duration, interstorm_duration)

        When each tuple pair is yielded, the grid scalar field 'rainfall__flux'
        is updated with the rainfall rate occuring during storm_duration.

        This generator method is designed for direct equivalence with the
        spatially resolved generators found elsewhere in Landlab.

        This generator will be useful in cases where the whole sequence of
        storms and interstorms doesn't need to be stored, where we can save
        memory this way.

        This method does not attempt to subdivide timesteps. If you want that,
        provide a delta_t at instantiation, and use
        yield_storm_interstorm_duration_intensity(subdivide_interstorms=True).

        The method will keep yielding until it reaches the RUN_TIME, where it
        will terminate. If it terminates during a storm, the final tuple will
        be (truncated_storm_duration, 0.). Otherwise it will be
        (storm_duration, truncated_interstorm_duration.)

        Yields
        ------
        tuple of float
            (storm_duration, interstorm_duration)

        Notes
        -----
        One recommended procedure is to instantiate the generator, then call
        instance.next() (in Python 2) or next(instance) (in Python 3)
        repeatedly to get the sequence (See Examples, below).

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> mg = RasterModelGrid((4, 5), (1., 1.))
        >>> precip = PrecipitationDistribution(mg, mean_storm_duration=1.5,
        ...     mean_interstorm_duration=15.0, mean_storm_depth=0.5,
        ...     total_t=46.)
        >>> storm_dts = []
        >>> interstorm_dts = []
        >>> intensities = []
        >>> precip.seed_generator(seedval=1)
        >>> for (storm_dt, interstorm_dt) in precip.yield_storms():
        ...     storm_dts.append(storm_dt)
        ...     interstorm_dts.append(interstorm_dt)
        ...     intensities.append(mg.at_grid['rainfall__flux'])
        >>> len(storm_dts) == 4  # 4 storms in the simulation
        True
        >>> len(interstorm_dts) == len(storm_dts)
        True
        >>> rf_intensities_to_test = np.array([0.8138257984406472,
        ...                                    0.15929112025199238,
        ...                                    0.17254519305000884,
        ...                                    0.09817611240558813])
        >>> np.allclose(intensities, rf_intensities_to_test)
        True
        >>> np.isclose(sum(storm_dts) + sum(interstorm_dts), 46.)  # total_t
        True
        >>> np.isclose(interstorm_dts[-1], 0.)  # sequence truncated
        True

        An alternative way to use the generator might be:

        >>> from six import next
        >>> # ^^this lets you "manually" get the next item from the iterator
        >>> _ = mg.at_grid.pop('rainfall__flux')  # remove the existing field
        >>> precip = PrecipitationDistribution(mg, mean_storm_duration=1.5,
        ...     mean_interstorm_duration=15.0, mean_storm_depth=0.5,
        ...     total_t=46.)
        >>> precip.seed_generator(seedval=1)
        >>> mystorm_generator = precip.yield_storms()
        >>> my_list_of_storms = []
        >>> for i in range(4):
        ...     my_list_of_storms.append(next(mystorm_generator))

        Note that an exception is thrown if you go too far:

        >>> my_list_of_storms.append(next(mystorm_generator))  # the 5th iter
        Traceback (most recent call last):
          ...
        StopIteration

        Also note that the generator won't terminate if you try this without
        first instantiating the generator:

        >>> allmytimes = []
        >>> for i in range(20):  # this will run just fine
        ...     allmytimes.append(next(precip.yield_storms()))
        >>> total_t = sum([sum(storm) for storm in allmytimes])
        >>> total_t > 46.
        True

        """
        # we must have instantiated with a grid, so check:
        assert hasattr(self, '_grid')

        # now exploit the existing generator to make this easier & less
        # redundant:
        delta_t = self.delta_t
        self.delta_t = None  # this is necessary to suppress chunking behaviour
        # in the other generator
        othergen = self.yield_storm_interstorm_duration_intensity()
        # enter a loop, to break as needed:
        tobreak = False
        while not tobreak:
            # we always start with a storm, so:
            try:
                (storm_dur, storm_int) = next(othergen)
            except StopIteration:
                break  # stop dead. We terminated at a good place
            try:
                (interstorm_dur, _) = next(othergen)
            except StopIteration:
                tobreak = True
                interstorm_dur = 0.
            # reset the rainfall__flux field, that got overstamped in the
            # interstorm iter:
            self.grid.at_grid['rainfall__flux'] = storm_int
            yield (storm_dur, interstorm_dur)
        # now, just in case, restore self.delta_t:
        self.delta_t = delta_t

    def generate_from_stretched_exponential(self, scale, shape):
        """Generate and return a random variable from a stretched exponential
        distribution with given scale and shape.

        Examples
        --------
        >>> np.random.seed(0)
        >>> np.round(np.random.rand(3), 6)  # these are our 3 rand #s to test
        array([ 0.548814,  0.715189,  0.602763])
        >>> from landlab.components import PrecipitationDistribution
        >>> pd = PrecipitationDistribution(mean_storm_duration=1.0,
        ...                                mean_interstorm_duration=1.0,
        ...                                mean_storm_depth=1.0)
        >>> np.random.seed(0)  # re-set seed so we get the same 3 #s
        >>> np.round(1000 * pd.generate_from_stretched_exponential(2.0, 0.5))
        720.0
        >>> np.round(1000 * pd.generate_from_stretched_exponential(2.0, 0.5))
        225.0
        >>> np.round(1000 * pd.generate_from_stretched_exponential(2.0, 0.5))
        513.0
        """
        return scale * ((-np.log(np.random.rand())) ** (1.0 / shape))

    def seed_generator(self, seedval=0):
        """Seed the random-number generator.

        The examples illustrate:
        (1) that we can get the same sequence again by re-seeding with the
            same value (the default is zero)
        (2) when we use a value other than the default, we get a different
            sequence

        Examples
        --------
        >>> precip = PrecipitationDistribution(mean_storm_duration=1.5,
        ...     mean_interstorm_duration=15.0, mean_storm_depth=0.5,
        ...     total_t=100.0, delta_t=1.)
        >>> round(precip.storm_duration, 2)
        2.79
        >>> round(precip.interstorm_duration, 2)
        21.28
        >>> round(precip.storm_depth, 2)
        2.45
        >>> round(precip.intensity, 2)
        0.88
        >>> precip.seed_generator() # re-seed and get same sequence again
        >>> round(precip.get_precipitation_event_duration(), 2)
        2.79
        >>> round(precip.get_interstorm_event_duration(), 2)
        21.28
        >>> round(precip.get_storm_depth(), 2)
        2.45
        >>> round(precip.get_storm_intensity(), 2)
        0.88
        >>> precip = PrecipitationDistribution(mean_storm_duration=1.5,
        ...     mean_interstorm_duration=15.0, mean_storm_depth=0.5,
        ...     total_t=100.0, delta_t=1., random_seed=1)
        >>> round(precip.storm_duration, 2) # diff't vals with diff't seed
        0.22
        >>> round(precip.interstorm_duration, 2)
        28.2
        >>> round(precip.storm_depth, 4)
        0.0012
        >>> round(precip.intensity, 4)
        0.0054
        """
        random.seed(seedval)
        np.random.seed(seedval)

    @property
    def elapsed_time(self):
        """Get the elapsed time recorded by the module.

        This will be particularly useful in the midst of a yield loop.
        """
        return self._elapsed_time

    @property
    def intensity(self):
        """Get the intensity of the most recent storm simulated.
        """
        return self.get_storm_intensity()



from .chi_index import ChiFinder
from .diffusion import LinearDiffuser
from .fire_generator import FireGenerator
from .detachment_ltd_erosion import DetachmentLtdErosion, DepthSlopeProductErosion
from .flexure import Flexure
from .flow_routing import FlowRouter, DepressionFinderAndRouter
from .nonlinear_diffusion import PerronNLDiffuse
from .flow_director import FlowDirectorD8
from .flow_director import FlowDirectorSteepest
from .flow_director import FlowDirectorMFD
from .flow_director import FlowDirectorDINF
from .flow_accum import FlowAccumulator
from .overland_flow import OverlandFlowBates, OverlandFlow
from .overland_flow import KinwaveImplicitOverlandFlow, KinwaveOverlandFlowModel
from .potentiality_flowrouting import PotentialityFlowRouter
from .pet import PotentialEvapotranspiration
from .radiation import Radiation
from .soil_moisture import SoilMoisture
from .vegetation_dynamics import Vegetation
from .sink_fill import SinkFiller
from .steepness_index import SteepnessFinder
from .stream_power import StreamPowerEroder, FastscapeEroder, StreamPowerSmoothThresholdEroder, SedDepEroder
from .uniform_precip import PrecipitationDistribution
from .soil_moisture import SoilInfiltrationGreenAmpt
from .plant_competition_ca import VegCA
from .gflex import gFlex
from .drainage_density import DrainageDensity
from .weathering import ExponentialWeatherer
from .depth_dependent_diffusion import DepthDependentDiffuser
from .taylor_nonlinear_hillslope_flux import TaylorNonLinearDiffuser
from .depth_dependent_taylor_soil_creep import DepthDependentTaylorDiffuser
from .erosion_deposition import ErosionDeposition
from .space import Space
from .landslides import LandslideProbability
from .transport_length_diffusion import TransportLengthHillslopeDiffuser
from .normal_fault import NormalFault

COMPONENTS = [ChiFinder, LinearDiffuser,
              Flexure, FlowRouter, DepressionFinderAndRouter,
              PerronNLDiffuse, OverlandFlowBates, OverlandFlow,
              KinwaveImplicitOverlandFlow,
              PotentialEvapotranspiration, PotentialityFlowRouter,
              Radiation, SinkFiller,
              StreamPowerEroder, StreamPowerSmoothThresholdEroder,
              FastscapeEroder, SedDepEroder,
              PrecipitationDistribution,
              SteepnessFinder, DetachmentLtdErosion, gFlex,
              SoilInfiltrationGreenAmpt, FireGenerator,
              SoilMoisture, Vegetation, VegCA, DrainageDensity,
              ExponentialWeatherer, DepthDependentDiffuser,
              TaylorNonLinearDiffuser, DepthSlopeProductErosion,
              FlowDirectorD8, FlowDirectorSteepest, FlowDirectorMFD,
              FlowDirectorDINF, FlowAccumulator, Space, ErosionDeposition,
              LandslideProbability, DepthDependentTaylorDiffuser,
              NormalFault]

__all__ = [cls.__name__ for cls in COMPONENTS]



#! /usr/bin/env python
"""
Unit tests for landlab.collections
"""
import pytest
from landlab import Palette, Arena, NoProvidersError

from landlab import Implements, ImplementsOrRaise
from landlab.framework.interfaces import BmiBase, BmiNoGrid


@Implements(BmiBase)
class Sample1(object):

    """A sample component."""

    __implements__ = (BmiBase, BmiNoGrid)

    _input_var_names = ["air__temperature", "surface__elevation"]
    _output_var_names = ["deposition__rate"]

    model_name = "Sample 1"
    author_name = "Eric Hutton"
    version = "0.1"
    time_units = "s"
    time_step_type = "fixed"
    step_method = "explicit"
    grid_type = "none"

    _vars = {"deposition__rate": [1.]}

    def initialize(self, name):
        pass

    def update(self):
        pass

    def finalize(self):
        pass

    def get_input_var_names(self):
        return self._input_var_names

    def get_output_var_names(self):
        return self._output_var_names

    def get_var_rank(self, name):
        return 0

    def get_start_time(self):
        return 0.

    def get_current_time(self):
        return 0.

    def get_end_time(self):
        return 100.

    def get_time_step(self):
        return 1.

    def get_var_type(self, name):
        return "float64"

    def get_var_units(self, name):
        return "m"

    def set_value(self, name, value):
        pass

    def get_value(self, name):
        return self._vars[name]


@ImplementsOrRaise(BmiBase)
class Sample2(object):

    """A sample component."""

    __implements__ = (BmiBase, BmiNoGrid)

    _input_var_names = ["deposition__rate"]
    _output_var_names = ["air__temperature", "surface__elevation"]

    model_name = "Sample 2"
    author_name = "Eric Hutton"
    version = "0.1"
    time_units = "s"
    time_step_type = "fixed"
    step_method = "explicit"
    grid_type = "none"

    _vars = {"air__temperature": [1.], "surface__elevation": [1.]}

    def initialize(self, name):
        pass

    def update(self):
        pass

    def finalize(self):
        pass

    def get_input_var_names(self):
        return self._input_var_names

    def get_output_var_names(self):
        return self._output_var_names

    def get_var_rank(self, name):
        return 0

    def get_start_time(self):
        return 0.

    def get_current_time(self):
        return 0.

    def get_end_time(self):
        return 100.

    def get_time_step(self):
        return 1.

    def get_var_type(self, name):
        return "float64"

    def get_var_units(self, name):
        return "m"

    def get_value(self, name):
        return self._vars[name]

    def set_value(self, name, value):
        pass


def test_empty_palette():
    """Create a palette without components."""
    palette = Palette()

    assert len(palette) == 0
    assert list(palette.list()) == []
    assert list(palette.keys()) == []
    assert palette.uses() == []
    assert palette.provides() == []

    providers = palette.find_provider("air__temperature")
    assert providers == []

    users = palette.find_user("air__temperature")
    assert users == []

    connections = palette.find_connections()
    assert connections == {}


def test_1_component_create():
    palette = Palette(sample=Sample1)

    assert len(palette) == 1


def test_1_component_dict_interface():
    palette = Palette(sample=Sample1)

    assert dict(sample=Sample1) == palette
    assert len(palette) == 1
    assert list(palette.keys()) == ["sample"]
    assert list(palette.values()) == [Sample1]

    items = list(palette.items())
    assert ("sample", Sample1) == items[0]


def test_1_component_list():
    palette = Palette(sample=Sample1)

    assert ["sample"] == list(palette.list())


def test_1_component_uses():
    palette = Palette(sample=Sample1)

    uses = palette.uses()
    uses.sort()
    assert ["air__temperature", "surface__elevation"] == uses


def test_1_component_provides():
    palette = Palette(sample=Sample1)

    provides = palette.provides()
    provides.sort()
    assert ["deposition__rate"] == provides


def test_1_component_find_providers():
    palette = Palette(sample=Sample1)

    providers = palette.find_provider("air__temperature")
    assert providers == []

    providers = palette.find_provider("deposition__rate")
    assert providers == ["sample"]


def test_1_component_find_users():
    palette = Palette(sample=Sample1)

    users = palette.find_user("air__temperature")
    assert users == ["sample"]


def test_1_component_find_connections():
    palette = Palette(sample=Sample1)

    with pytest.raises(NoProvidersError):
        palette.find_connections()


def test_2_components_create():
    palette = Palette(one=Sample1, two=Sample2)

    assert len(palette) == 2


def test_2_components_dict_interface():
    palette = Palette(one=Sample1, two=Sample2)

    assert dict(one=Sample1, two=Sample2) == palette
    assert len(palette) == 2

    keys = list(palette.keys())
    keys.sort()
    assert ["one", "two"] == keys

    values = palette.values()
    assert 2 == len(values)
    assert Sample1 in values and Sample2 in values

    items = list(palette.items())
    items.sort()
    assert 2 == len(items)
    assert ("one", Sample1) == items[0]
    assert ("two", Sample2) == items[1]


def test_2_components_list():
    palette = Palette(one=Sample1, two=Sample2)

    components = list(palette.list())
    components.sort()
    assert ["one", "two"] == components


def test_2_components_uses():
    palette = Palette(one=Sample1, two=Sample2)

    uses = palette.uses()
    uses.sort()
    assert ["air__temperature", "deposition__rate", "surface__elevation"] == uses


def test_2_components_provides():
    palette = Palette(one=Sample1, two=Sample2)

    provides = palette.provides()
    provides.sort()
    assert ["air__temperature", "deposition__rate", "surface__elevation"] == provides


def test_2_components_find_providers():
    palette = Palette(one=Sample1, two=Sample2)

    providers = palette.find_provider("air__temperature")
    assert ["two"] == providers

    providers = palette.find_provider("deposition__rate")
    assert ["one"] == providers


def test_2_components_find_users():
    palette = Palette(one=Sample1, two=Sample2)

    users = palette.find_user("air__temperature")
    assert ["one"] == users


def test_2_components_find_connections():
    palette = Palette(one=Sample1, two=Sample2)

    connections = {
        "one": {"deposition__rate": ["two"]},
        "two": {"air__temperature": ["one"], "surface__elevation": ["one"]},
    }
    assert connections == palette.find_connections()


def test_arena_instantiate():
    arena = Arena()
    assert dict() == arena

    arena.instantiate(Sample1, "one")
    assert 1 == len(arena)
    assert "one" in arena
    assert isinstance(arena["one"], Sample1)

    arena.instantiate(Sample2, "two")
    assert 2 == len(arena)
    assert "one" in arena and "two" in arena
    assert isinstance(arena["one"], Sample1)
    assert isinstance(arena["two"], Sample2)


def test_arena_connect():
    arena = Arena()
    arena.instantiate(Sample1, "one")
    arena.instantiate(Sample2, "two")

    arena.connect("one", "two", "air__temperature")
    arena.connect("one", "two", "surface__elevation")
    arena.connect("two", "one", "deposition__rate")

    dz = arena["one"].get_value("deposition__rate")
    t = arena["two"].get_value("air__temperature")
    z = arena["two"].get_value("surface__elevation")


def test_arena_walk():
    arena = Arena()
    arena.instantiate(Sample1, "one")
    arena.instantiate(Sample2, "two")

    arena.connect("one", "two", "air__temperature")
    arena.connect("one", "two", "surface__elevation")
    arena.connect("two", "one", "deposition__rate")

    tree = arena.walk("one")
    assert ["one", "two"] == tree



#! /usr/env/python
"""

Component that models 2D diffusion using an explicit finite-volume method.

Created July 2013 GT
Last updated March 2016 DEJH with LL v1.0 component style
"""

from __future__ import print_function

import numpy as np
from six.moves import range
from inspect import getmro

from landlab import ModelParameterDictionary, Component, FieldError, \
    create_and_initialize_grid, FIXED_GRADIENT_BOUNDARY, FIXED_LINK, \
    RasterModelGrid, INACTIVE_LINK
from landlab.core.model_parameter_dictionary import MissingKeyError
from landlab.utils.decorators import use_file_name_or_kwds

_ALPHA = 0.15   # time-step stability factor
# ^0.25 not restrictive enough at meter scales w S~1 (possible cases)


class LinearDiffuser(Component):
    """
    This component implements linear diffusion of a Landlab field.

    Component assumes grid does not deform. If the boundary conditions on the
    grid change after component instantiation, be sure to also call
    :func:`updated_boundary_conditions` to ensure these are reflected in the
    component (especially if fixed_links are present).

    The method keyword allows control of the way the solver works. Options
    other than 'simple' will make the component run slower, but give second
    order solutions that incorporate information from more distal nodes (i.e.,
    the diagonals). Note that the option 'resolve_on_patches' can result in
    somewhat counterintuitive behaviour - this option tells the component to
    treat the diffusivity as a field **with directionality to it** (i.e., that
    the diffusivites are defined on links). Thus if all links have the same
    diffusivity value, with this flag active "effective" diffusivities
    at the nodes will be *higher* than this value (by a factor of root 2) as
    the diffusivity at each patch will be the mean vector sum of that at the
    bounding links.

    The primary method of this class is :func:`run_one_step`.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> import numpy as np
    >>> mg = RasterModelGrid((9, 9), 1.)
    >>> z = mg.add_zeros('node', 'topographic__elevation')
    >>> z.reshape((9, 9))[4, 4] = 1.
    >>> mg.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> ld = LinearDiffuser(mg, linear_diffusivity=1.)
    >>> for i in range(1):
    ...     ld.run_one_step(1.)
    >>> np.isclose(z[mg.core_nodes].sum(), 1.)
    True
    >>> mg2 = RasterModelGrid((5, 30), 1.)
    >>> z2 = mg2.add_zeros('node', 'topographic__elevation')
    >>> z2.reshape((5, 30))[2, 8] = 1.
    >>> z2.reshape((5, 30))[2, 22] = 1.
    >>> mg2.set_closed_boundaries_at_grid_edges(True, True, True, True)
    >>> kd = mg2.node_x/mg2.node_x.mean()
    >>> ld2 = LinearDiffuser(mg2, linear_diffusivity=kd)
    >>> for i in range(10):
    ...     ld2.run_one_step(0.1)
    >>> z2[mg2.core_nodes].sum() == 2.
    True
    >>> z2.reshape((5, 30))[2, 8] > z2.reshape((5, 30))[2, 22]
    True

    An example using links:

    >>> mg1 = RasterModelGrid((10, 10), 100.)
    >>> mg2 = RasterModelGrid((10, 10), 100.)
    >>> z1 = mg1.add_zeros('node', 'topographic__elevation')
    >>> z2 = mg2.add_zeros('node', 'topographic__elevation')
    >>> dt = 1.
    >>> nt = 10
    >>> kappa_links = mg2.add_ones('link', 'surface_water__discharge')
    >>> kappa_links *= 10000.
    >>> dfn1 = LinearDiffuser(mg1, linear_diffusivity=10000.)
    >>> dfn2 = LinearDiffuser(mg2, linear_diffusivity='surface_water__discharge')
    >>> for i in range(nt):
    ...     z1[mg1.core_nodes] += 1.
    ...     z2[mg2.core_nodes] += 1.
    ...     dfn1.run_one_step(dt)
    ...     dfn2.run_one_step(dt)
    >>> np.allclose(z1, z2)
    True
    >>> z2.fill(0.)
    >>> dfn2 = LinearDiffuser(mg2, linear_diffusivity='surface_water__discharge',
    ...                       method='resolve_on_patches')
    >>> for i in range(nt):
    ...     z2[mg2.core_nodes] += 1.
    ...     dfn2.run_one_step(dt)
    >>> np.all(z2[mg2.core_nodes] < z1[mg2.core_nodes])
    True
    """

    _name = 'LinearDiffuser'

    _input_var_names = ('topographic__elevation',)

    _output_var_names = ('topographic__elevation',
                         'topographic__gradient',
                         'hillslope_sediment__unit_volume_flux',
                         )

    _var_units = {'topographic__elevation': 'm',
                  'topographic__gradient': '-',
                  'hillslope_sediment__unit_volume_flux': 'm**2/s',
                  }

    _var_mapping = {'topographic__elevation': 'node',
                    'topographic__gradient': 'link',
                    'hillslope_sediment__unit_volume_flux': 'link',
                    }

    _var_doc = {
        'topographic__elevation': ('Land surface topographic elevation; can ' +
                                   'be overwritten in initialization'),
        'topographic__gradient': 'Gradient of surface, on links',
        'hillslope_sediment__unit_volume_flux': 'Volume flux per unit width along links'
    }

    @use_file_name_or_kwds
    def __init__(self, grid, linear_diffusivity=None, method='simple',
                 deposit=True, **kwds):
        """
        Parameters
        ----------
        grid : ModelGrid
            A grid.
        linear_diffusivity : float, array, or field name (m**2/time)
            The diffusivity. If an array or field name, these must be the
            diffusivities on either nodes or links - the component will
            distinguish which based on array length. Values on nodes will be
            mapped to links using an upwind scheme in the simple case.
        method : {'simple', 'resolve_on_patches', 'on_diagonals'}
            The method used to represent the fluxes. 'simple' solves a finite
            difference method with a simple staggered grid scheme onto the links.
            'resolve_on_patches' solves the scheme by mapping both slopes and
            diffusivities onto the patches and solving there before resolving
            values back to the nodes (and at the moment requires a raster grid).
            Note that this scheme enforces directionality in the diffusion field;
            it's no longer just a scalar field. Thus diffusivities must be defined
            *on links* when this option is chosen.
            'on_diagonals' permits Raster diagonals to carry the diffusional
            fluxes. These latter techniques are more computationally expensive,
            but can suppress cardinal direction artifacts if diffusion is
            performed on a raster. 'on_diagonals' pretends that the "faces" of a
            cell with 8 links are represented by a stretched regular octagon set
            within the true cell.
        deposit : {True, False}
            Whether diffusive material can be deposited. True means that diffusive 
            material will be deposited if the divergence of sediment flux is 
            negative. False means that even when the divergence of sediment flux is 
            negative, no material is deposited. (No deposition ever.) The False
            case is a bit of a band-aid to account for cases when fluvial incision
            likely removes any material that would be deposited. If one couples
            fluvial detachment-limited incision with linear diffusion, the channels
            will not reach the predicted analytical solution unless deposit is set
            to False.
        """
        self._grid = grid
        self._bc_set_code = self.grid.bc_set_code
        assert method in ('simple', 'resolve_on_patches', 'on_diagonals')
        if method == 'resolve_on_patches':
            assert isinstance(self.grid, RasterModelGrid)
            self._use_patches = True
        else:
            self._use_patches = False
        if method == 'on_diagonals' and isinstance(self.grid, RasterModelGrid):
            self._use_diags = True
        else:
            self._use_diags = False
        self.current_time = 0.
        self._run_before = False
        self._kd_on_links = False
        if linear_diffusivity is not None:
            if type(linear_diffusivity) is not str:
                self._kd = linear_diffusivity
                if type(self._kd) in (float, int):
                    if type(self._kd) is int:
                        self._kd = float(self._kd)
                else:
                    if self._kd.size == self.grid.number_of_links:
                        self._kd_on_links = True
                    else:
                        assert self._kd.size == self.grid.number_of_nodes
            else:
                try:
                    self._kd = self.grid.at_link[linear_diffusivity]
                    self._kd_on_links = True
                except KeyError:
                    self._kd = self.grid.at_node[linear_diffusivity]
        else:
            raise KeyError("linear_diffusivity must be provided to the " +
                           "LinearDiffuser component")
        if self._kd_on_links is True:
            assert isinstance(self.grid, RasterModelGrid)

        # if we're using patches, it is VITAL that diffusivity is defined on
        # links. The whole point of this functionality is that we honour
        # *directionality* in the diffusivities.
        if self._use_patches:
            assert self._kd_on_links
        # set _deposit flag to tell code whether or not diffusion can deposit.    
        
        self._deposit = deposit
    
        # for component back compatibility (undocumented):
        # note component can NO LONGER do internal uplift, at all.
        # ###
        self.timestep_in = kwds.pop('dt', None)
        if 'values_to_diffuse' in kwds.keys():
            self.values_to_diffuse = kwds.pop('values_to_diffuse')
            for mytups in (self._input_var_names, self._output_var_names):
                myset = set(mytups)
                myset.remove('topographic__elevation')
                myset.add(self.values_to_diffuse)
                mytups = tuple(myset)
            for mydicts in (self._var_units, self._var_mapping, self._var_doc):
                mydicts[self.values_to_diffuse] = mydicts.pop(
                    'topographic__elevation')
        else:
            self.values_to_diffuse = 'topographic__elevation'
        # Raise an error if somehow someone is using this weird functionality
        if self._grid is None:
            raise ValueError('You must now provide an existing grid!')
        # ###

        # Set internal time step
        # ..todo:
        #   implement mechanism to compute time-steps dynamically if grid is
        #   adaptive/changing
        # as of modern componentization (Spring '16), this can take arrays
        # and irregular grids
        # CFL condition precalc:
        CFL_prefactor = _ALPHA * self.grid.length_of_link[
            :self.grid.number_of_links] ** 2.
        # ^ link_length can include diags, if not careful...
        self._CFL_actives_prefactor = CFL_prefactor[self.grid.active_links]
        # ^note we can do this as topology shouldn't be changing

        # Get a list of interior cells
        self.interior_cells = self.grid.node_at_core_cell

        self.z = self.grid.at_node[self.values_to_diffuse]
        self.dqsds = self.grid.zeros('node', dtype=float)
        if not self._use_diags:
            g = self.grid.zeros(at='link')
            qs = self.grid.zeros(at='link')
            try:
                self.g = self.grid.add_field('link', 'topographic__gradient',
                                             g, noclobber=True)
                # ^note this will object if this exists already
            except FieldError:  # keep a ref
                self.g = self.grid.at_link['topographic__gradient']
            try:
                self.qs = self.grid.add_field('link', 'hillslope_sediment__unit_volume_flux', qs,
                                              noclobber=True)
            except FieldError:
                self.qs = self.grid.at_link['hillslope_sediment__unit_volume_flux']
            # note all these terms are deliberately loose, as we won't always
            # be dealing with topo
        else:
            g = np.zeros(self.grid.number_of_d8, dtype=float)
            qs = np.zeros(self.grid.number_of_d8, dtype=float)
            self.g = g
            self.qs = qs
            # now we have to choose what the face width of a diagonal is...
            # Adopt a regular octagon config if it's a square raster, and
            # stretch this geometry as needed.
            # Conceptually, this means we're passing mass across diamond-
            # shaped holes centered at the corners.
            # Note that this WON'T affect the inferred cell size - that's
            # still derived from the rectangle.
            self._d8width_face_at_link = np.empty(
                self.grid.number_of_d8)
            # note there will be null entries here
            # by our defs, every active link must have a face.
            # calc the length of a diag "face":
            rt2 = np.sqrt(2.)
            horizontal_face = self.grid.dx / (1. + rt2)
            vertical_face = self.grid.dy / (1. + rt2)
            diag_face = np.sqrt(0.5 * (
                horizontal_face ** 2 + vertical_face ** 2))
            self._hoz = self.grid.horizontal_links.flatten()
            self._vert = self.grid.vertical_links.flatten()
            self._d8width_face_at_link[self._hoz] = vertical_face
            self._d8width_face_at_link[self._vert] = horizontal_face
            # ^ this operation pastes in faces where there are none, but
            # we'll never use them
            self._d8width_face_at_link[self.grid.number_of_links:] = diag_face

        self._vertlinkcomp = np.sin(self.grid.angle_of_link)
        self._hozlinkcomp = np.cos(self.grid.angle_of_link)

        if self._use_patches or self._kd_on_links:
            mg = self.grid
            try:
                self._hoz = self.grid.horizontal_links.flatten()
                self._vert = self.grid.vertical_links.flatten()
            except AttributeError:
                pass
            self._x_link_patches = mg.patches_at_link[self._hoz]
            x_link_patch_pres = mg.patches_present_at_link[self._hoz]
            self._x_link_patch_mask = np.logical_not(x_link_patch_pres)
            self._y_link_patches = mg.patches_at_link[self._vert]
            y_link_patch_pres = mg.patches_present_at_link[self._vert]
            self._y_link_patch_mask = np.logical_not(y_link_patch_pres)
            self._hoz_link_neighbors = np.empty((self._hoz.size, 4),
                                                dtype=int)
            self._vert_link_neighbors = np.empty((self._vert.size, 4),
                                                 dtype=int)

        # do some pre-work to make fixed grad BC updating faster in the loop:
        self.updated_boundary_conditions()

    def updated_boundary_conditions(self):
        """Call if grid BCs are updated after component instantiation.

        Sets `fixed_grad_nodes`, `fixed_grad_anchors`, & `fixed_grad_offsets`,
        such that::

            value[fixed_grad_nodes] = value[fixed_grad_anchors] + offset

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> import numpy as np
        >>> mg = RasterModelGrid((4, 5), 1.)
        >>> z = mg.add_zeros('node', 'topographic__elevation')
        >>> z[mg.core_nodes] = 1.
        >>> ld = LinearDiffuser(mg, linear_diffusivity=1.)
        >>> ld.fixed_grad_nodes.size == 0
        True
        >>> ld.fixed_grad_anchors.size == 0
        True
        >>> ld.fixed_grad_offsets.size == 0
        True
        >>> mg.at_link['topographic__slope'] = mg.calc_grad_at_link(
        ...     'topographic__elevation')
        >>> mg.set_fixed_link_boundaries_at_grid_edges(True, True, True, True)
        >>> ld.updated_boundary_conditions()
        >>> ld.fixed_grad_nodes
        array([ 1,  2,  3,  5,  9, 10, 14, 16, 17, 18])
        >>> ld.fixed_grad_anchors
        array([ 6,  7,  8,  6,  8, 11, 13, 11, 12, 13])
        >>> ld.fixed_grad_offsets
        array([-1., -1., -1., -1., -1., -1., -1., -1., -1., -1.])
        >>> np.allclose(z[ld.fixed_grad_nodes],
        ...             z[ld.fixed_grad_anchors] + ld.fixed_grad_offsets)
        True
        """
        fixed_grad_nodes = np.where(self.grid.status_at_node ==
                                    FIXED_GRADIENT_BOUNDARY)[0]
        heads = self.grid.node_at_link_head[self.grid.fixed_links]
        tails = self.grid.node_at_link_tail[self.grid.fixed_links]
        head_is_fixed = np.in1d(heads, fixed_grad_nodes)
        self.fixed_grad_nodes = np.where(head_is_fixed, heads, tails)
        self.fixed_grad_anchors = np.where(head_is_fixed, tails, heads)
        vals = self.grid.at_node[self.values_to_diffuse]
        self.fixed_grad_offsets = (vals[self.fixed_grad_nodes] -
                                   vals[self.fixed_grad_anchors])
        if self._use_diags:
            self.g.fill(0.)

        if self._kd_on_links or self._use_patches:
            mg = self.grid
            x_link_patch_pres = mg.patches_present_at_link[self._hoz]
            self._x_link_patch_mask = np.logical_not(x_link_patch_pres)
            y_link_patch_pres = mg.patches_present_at_link[self._vert]
            self._y_link_patch_mask = np.logical_not(y_link_patch_pres)
            self._hoz_link_neighbors[:, :2] = mg.links_at_node[
                mg.node_at_link_head[self._hoz], 1:4:2]
            self._hoz_link_neighbors[:, 2:] = mg.links_at_node[
                mg.node_at_link_tail[self._hoz], 1:4:2]
            self._vert_link_neighbors[:, :2] = mg.links_at_node[
                mg.node_at_link_head[self._vert], 0:3:2]
            self._vert_link_neighbors[:, 2:] = mg.links_at_node[
                mg.node_at_link_tail[self._vert], 0:3:2]
            self._vert_link_badlinks = np.logical_or(
                mg.status_at_link[self._vert_link_neighbors] == INACTIVE_LINK,
                self._vert_link_neighbors == -1)
            self._hoz_link_badlinks = np.logical_or(
                mg.status_at_link[self._hoz_link_neighbors] == INACTIVE_LINK,
                self._hoz_link_neighbors == -1)

    def diffuse(self, dt, **kwds):
        """
        See :func:`run_one_step`.
        """
        if 'internal_uplift' in kwds.keys():
            raise KeyError('LinearDiffuser can no longer work with internal ' +
                           'uplift')
        mg = self.grid
        z = self.grid.at_node[self.values_to_diffuse]

        if not self._run_before:
            self.updated_boundary_conditions()  # just in case
            self._run_before = True
        if self._bc_set_code != self.grid.bc_set_code:
            self.updated_boundary_conditions()
            self._bc_set_code = self.grid.bc_set_code

        core_nodes = self.grid.node_at_core_cell
        # do mapping of array kd here, in case it points at an updating
        # field:
        if type(self._kd) is np.ndarray:
            if not self._kd_on_links:
                kd_links = self.grid.map_max_of_link_nodes_to_link(self._kd)
                kd_activelinks = kd_links[self.grid.active_links]
                # re-derive CFL condition, as could change dynamically:
                dt_links = self._CFL_actives_prefactor / kd_activelinks
                self.dt = np.nanmin(dt_links)
            else:
                kd_links = self._kd
                kd_activelinks = self._kd[self.grid.active_links]
                dt_links = self._CFL_actives_prefactor / kd_activelinks
                self.dt_links = dt_links
                self.dt = np.nanmin(np.fabs(dt_links))
        else:
            kd_activelinks = self._kd
            # re-derive CFL condition, as could change dynamically:
            dt_links = self._CFL_actives_prefactor / kd_activelinks
            self.dt = np.nanmin(dt_links)

        if self._use_patches:
            # need this else diffusivities on inactive links deform off-angle
            # calculations
            kd_links = kd_links.copy()
            kd_links[self.grid.status_at_link == INACTIVE_LINK] = 0.

        # Take the smaller of delt or built-in time-step size self.dt
        self.tstep_ratio = dt / self.dt
        repeats = int(self.tstep_ratio // 1.)
        extra_time = self.tstep_ratio - repeats

        # Can really get into trouble if no diffusivity happens but we run...
        if self.dt < np.inf:
            loops = repeats+1
        else:
            loops = 0
        for i in range(loops):
            if not self._use_diags:
                grads = mg.calc_grad_at_link(z)
                self.g[mg.active_links] = grads[mg.active_links]
                if not self._use_patches:  # currently forbidden
                    # if diffusivity is an array, self._kd is already
                    # active_links-long
                    self.qs[mg.active_links] = (
                        -kd_activelinks * self.g[mg.active_links])
                    # Calculate the net deposition/erosion rate at each node
                    mg.calc_flux_div_at_node(self.qs, out=self.dqsds)
                else:  # project onto patches
                    slx = mg.zeros('link')
                    sly = mg.zeros('link')
                    slx[self._hoz] = self.g[self._hoz]
                    sly[self._vert] = self.g[self._vert]
                    patch_dx, patch_dy = mg.calc_grad_at_patch(z)
                    xvecs_vert = np.ma.array(patch_dx[self._y_link_patches],
                                             mask=self._y_link_patch_mask)
                    slx[self._vert] = xvecs_vert.mean()
                    yvecs_hoz = np.ma.array(patch_dy[self._x_link_patches],
                                            mask=self._x_link_patch_mask)
                    sly[self._hoz] = yvecs_hoz.mean()
                    # now map diffusivities (already on links, but we want
                    # more spatial averaging)
                    Kx = mg.zeros('link')
                    Ky = mg.zeros('link')
                    Kx[self._hoz] = kd_links[self._hoz]
                    Ky[self._vert] = kd_links[self._vert]
                    vert_link_crosslink_K = np.ma.array(
                        kd_links[self._vert_link_neighbors],
                        mask=self._vert_link_badlinks)
                    hoz_link_crosslink_K = np.ma.array(
                        kd_links[self._hoz_link_neighbors],
                        mask=self._hoz_link_badlinks)
                    Kx[self._vert] = vert_link_crosslink_K.mean(axis=1)
                    Ky[self._hoz] = hoz_link_crosslink_K.mean(axis=1)
                    Cslope = np.sqrt(slx ** 2 + sly ** 2)
                    v = np.sqrt(Kx ** 2 + Ky ** 2)
                    flux_links = v * Cslope
                    # NEW, to resolve issue with K being off angle to S:
                    # in fact, no. Doing this just makes this equivalent
                    # to the basic diffuser, but with a bunch more crap
                    # involved.
                    # flux_x = slx * Kx
                    # flux_y = sly * Ky
                    # flux_links = np.sqrt(flux_x*flux_x + flux_y*flux_y)
                    theta = np.arctan(np.fabs(sly) / (np.fabs(slx) + 1.e-10))
                    flux_links[self._hoz] *= (
                        np.sign(slx[self._hoz]) * np.cos(theta[self._hoz]))
                    flux_links[self._vert] *= (
                        np.sign(sly[self._vert]) * np.sin(theta[self._vert]))
                    # zero out the inactive links
                    self.qs[mg.active_links] = -flux_links[mg.active_links]

                    self.grid.calc_flux_div_at_node(self.qs, out=self.dqsds)

            else:  # ..._use_diags
                # NB: this is dirty code. It uses the obsolete diagonal data
                # structures, and necessarily has to do a bunch of mapping
                # on the fly.

                # remap the kds onto the links, as necessary
                if isinstance(self._kd, np.ndarray):
                    d8link_kd = np.empty(self.grid.number_of_d8,
                                         dtype=float)
                    d8link_kd[self.grid.active_links] = kd_activelinks
                    d8link_kd[self.grid.active_diagonals] = np.amax(
                        self._kd[self.grid.nodes_at_diagonal[self.grid.active_diagonals]],
                                 axis=1).flatten()
                else:
                    d8link_kd = self._kd
                self.g[self.grid.active_links] = self.grid.calc_grad_at_link(
                    z)[self.grid.active_links]
                self.g[self.grid.active_diagonals] = ((
                    z[self.grid._diag_activelink_tonode] -
                    z[self.grid._diag_activelink_fromnode]) /
                        self.grid.length_of_d8[self.grid.active_diagonals])
                self.qs[:] = -d8link_kd * self.g

                total_flux = self.qs * self._d8width_face_at_link  # nlinks
                totalflux_allnodes = (
                    total_flux[self.grid.links_at_node] *
                    self.grid.active_link_dirs_at_node).sum(axis=1)
                totalflux_allnodes += (
                    total_flux[self.grid.d8s_at_node[:, 4:]] *
                    self.grid.active_diagonal_dirs_at_node).sum(axis=1)
                self.dqsds[self.grid.node_at_cell] = -totalflux_allnodes[
                    self.grid.node_at_cell] / self.grid.area_of_cell

            # Calculate the total rate of elevation change
            dzdt = - self.dqsds
            if not self._deposit :
                dzdt[np.where(dzdt>0)] = 0.
            # Update the elevations
            timestep = self.dt
            if i == (repeats):
                timestep *= extra_time
            else:
                pass
            self.grid.at_node[self.values_to_diffuse][core_nodes] += dzdt[
                core_nodes] * timestep

            # check the BCs, update if fixed gradient
            vals = self.grid.at_node[self.values_to_diffuse]
            vals[self.fixed_grad_nodes] = (vals[self.fixed_grad_anchors] +
                                           self.fixed_grad_offsets)

        return self.grid

    def run_one_step(self, dt, **kwds):
        """Run the diffuser for one timestep, dt.

        If the imposed timestep dt is longer than the Courant-Friedrichs-Lewy
        condition for the diffusion, this timestep will be internally divided
        as the component runs, as needed.

        Parameters
        ----------
        dt : float (time)
            The imposed timestep.
        """
        self.diffuse(dt, **kwds)

    @property
    def time_step(self):
        """Returns internal time-step size (as a property).
        """
        return self.dt



#! /usr/env/python
"""Fastscape stream power erosion."""

# This module attempts to "component-ify" GT's Fastscape stream
# power erosion.
# Created DEJH, March 2014.
from __future__ import print_function

from six import string_types

import numpy as np
from landlab import Component

from landlab.utils.decorators import use_file_name_or_kwds

from landlab import BAD_INDEX_VALUE as UNDEFINED_INDEX

from landlab import RasterModelGrid

from .cfuncs import (brent_method_erode_fixed_threshold,
                     brent_method_erode_variable_threshold)


class FastscapeEroder(Component):

    r"""Fastscape stream power erosion.

    This class uses the Braun-Willett Fastscape approach to calculate the
    amount of erosion at each node in a grid, following a stream power
    framework. This should allow it to be stable against larger timesteps
    than an explicit stream power scheme.

    Note that although this scheme is nominally implicit, and will reach a
    numerically-correct solution under topographic steady state regardless of
    timestep length, the accuracy of transient solutions is *not* timestep
    independent (see Braun & Willett 2013, Appendix B for further details).
    Although the scheme remains significantly more robust and permits longer
    timesteps than a traditional explicit solver under such conditions, it
    is still possible to create numerical instability through use of too long
    a timestep while using this component. The user is cautioned to check their
    implementation is behaving stably before fully trusting it.

    Stream power erosion is implemented as:

    .. math::

        E = K (\textit{rainfall_intensity} \, A) ^ m  S ^ n -
               \textit{threshold_sp}

    if :math:`K A ^ m S ^ n > \textit{threshold_sp}`, and:

    .. math:: E = 0,

    if :math:`K A^m S^n <= \textit{threshold_sp}`.

    This module assumes you have already run
    :func:`landlab.components.flow_routing.route_flow_dn.FlowRouter.route_flow`
    in the same timestep. It looks for 'flow__upstream_node_order',
    'flow__link_to_receiver_node', 'drainage_area', 'flow__receiver_node', and
    'topographic__elevation' at the nodes in the grid. 'drainage_area' should
    be in area upstream, not volume (i.e., set runoff_rate=1.0 when calling
    FlowRouter.route_flow).

    The primary method of this class is :func:`run_one_step`.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> from landlab import CLOSED_BOUNDARY, FIXED_VALUE_BOUNDARY
    >>> from landlab.components import FlowRouter, FastscapeEroder

    >>> grid = RasterModelGrid((5, 5), spacing=10.)
    >>> z = np.array([7.,  7.,  7.,  7.,  7.,
    ...               7.,  5., 3.2,  6.,  7.,
    ...               7.,  2.,  3.,  5.,  7.,
    ...               7.,  1., 1.9,  4.,  7.,
    ...               7.,  0.,  7.,  7.,  7.])
    >>> z = grid.add_field('topographic__elevation', z, at='node')
    >>> fr = FlowRouter(grid)
    >>> sp = FastscapeEroder(grid, K_sp=1.)
    >>> fr.run_one_step()
    >>> sp.run_one_step(dt=1.)
    >>> z  # doctest: +NORMALIZE_WHITESPACE
    array([ 7.        ,  7.        ,  7.        ,  7.        ,  7.        ,
            7.        ,  2.92996598,  2.02996598,  4.01498299,  7.        ,
            7.        ,  0.85993197,  1.87743897,  3.28268321,  7.        ,
            7.        ,  0.28989795,  0.85403051,  2.42701526,  7.        ,
            7.        ,  0.        ,  7.        ,  7.        ,  7.        ])

    >>> grid = RasterModelGrid((3, 7), spacing=1.)
    >>> z = np.array(grid.node_x ** 2.)
    >>> z = grid.add_field('topographic__elevation', z, at='node')
    >>> grid.status_at_node[grid.nodes_at_left_edge] = FIXED_VALUE_BOUNDARY
    >>> grid.status_at_node[grid.nodes_at_top_edge] = CLOSED_BOUNDARY
    >>> grid.status_at_node[grid.nodes_at_bottom_edge] = CLOSED_BOUNDARY
    >>> grid.status_at_node[grid.nodes_at_right_edge] = CLOSED_BOUNDARY
    >>> fr = FlowRouter(grid)
    >>> sp = FastscapeEroder(grid, K_sp=0.1, m_sp=0., n_sp=2.,
    ...                      threshold_sp=2.)
    >>> fr.run_one_step()
    >>> sp.run_one_step(dt=10.)
    >>> z.reshape(grid.shape)[1, :]  # doctest: +NORMALIZE_WHITESPACE
    array([  0.        ,   1.        ,   4.        ,   8.52493781,
            13.29039716,  18.44367965,  36.        ])

    >>> grid = RasterModelGrid((3, 7), spacing=1.)
    >>> z = np.array(grid.node_x ** 2.)
    >>> z = grid.add_field('topographic__elevation', z, at='node')
    >>> grid.status_at_node[grid.nodes_at_left_edge] = FIXED_VALUE_BOUNDARY
    >>> grid.status_at_node[grid.nodes_at_top_edge] = CLOSED_BOUNDARY
    >>> grid.status_at_node[grid.nodes_at_bottom_edge] = CLOSED_BOUNDARY
    >>> grid.status_at_node[grid.nodes_at_right_edge] = CLOSED_BOUNDARY
    >>> fr = FlowRouter(grid)
    >>> K_field = grid.ones(at='node') # K can be a field
    >>> sp = FastscapeEroder(grid, K_sp=K_field, m_sp=1., n_sp=0.6,
    ...                      threshold_sp=grid.node_x,
    ...                      rainfall_intensity=2.)
    >>> fr.run_one_step()
    >>> sp.run_one_step(1.)
    >>> z.reshape(grid.shape)[1, :]  # doctest: +NORMALIZE_WHITESPACE
    array([  0.        ,   0.0647484 ,   0.58634455,   2.67253503,
             8.49212152,  20.92606987,  36.        ])
    >>> previous_z = z.copy()
    >>> sp.run_one_step(1., rainfall_intensity_if_used=0.)
    >>> np.allclose(z, previous_z)
    True
    """

    _name = 'FastscapeEroder'

    _input_var_names = (
        'topographic__elevation',
        'drainage_area',
        'flow__link_to_receiver_node',
        'flow__upstream_node_order',
        'flow__receiver_node',
    )

    _output_var_names = (
        'topographic__elevation',
    )

    _var_units = {
        'topographic__elevation': 'm',
        'drainage_area': 'm**2',
        'flow__link_to_receiver_node': '-',
        'flow__upstream_node_order': '-',
        'flow__receiver_node': '-',
    }

    _var_mapping = {
        'topographic__elevation': 'node',
        'drainage_area': 'node',
        'flow__link_to_receiver_node': 'node',
        'flow__upstream_node_order': 'node',
        'flow__receiver_node': 'node',
    }

    _var_doc = {
        'topographic__elevation': 'Land surface topographic elevation',
        'drainage_area':
            "Upstream accumulated surface area contributing to the node's "
            "discharge",
        'flow__link_to_receiver_node':
            'ID of link downstream of each node, which carries the discharge',
        'flow__upstream_node_order':
            'Node array containing downstream-to-upstream ordered list of '
            'node IDs',
        'flow__receiver_node':
            'Node array of receivers (node that receives flow from current '
            'node)',
    }

    @use_file_name_or_kwds
    def __init__(self, grid, K_sp=None, m_sp=0.5, n_sp=1., threshold_sp=0.,
                 rainfall_intensity=1., discharge_name='drainage_area',
                 **kwds):
        """
        Initialize the Fastscape stream power component. Note: a timestep,
        dt, can no longer be supplied to this component through the input file.
        It must instead be passed directly to the run method.

        Parameters
        ----------
        grid : ModelGrid
            A grid.
        K_sp : float, array, or field name
            K in the stream power equation (units vary with other parameters).
        m_sp : float, optional
            m in the stream power equation (power on drainage area).
        n_sp : float, optional
            n in the stream power equation (power on slope).
        rainfall intensity : float, array, or field name; optional
            Modifying factor on drainage area to convert it to a true water
            volume flux in (m/time). i.e., E = K * (r_i*A)**m * S**n
        discharge_name : string; optional
            Name of field to use for discharge proxy. Defaults to 'drainage_area',
            which means the component will expect the driver or another component
            to have created and populated a 'drainage_area' field. To use a
            different field, such as 'surface_water__discharge', give its name in
            this argument.
        """
        self._grid = grid

        self.K = K_sp  # overwritten below in special cases
        self.m = float(m_sp)
        self.n = float(n_sp)
        if isinstance(threshold_sp, (float, int)):
            self.thresholds = float(threshold_sp)
        else:
            if isinstance(threshold_sp, string_types):
                self.thresholds = self.grid.at_node[threshold_sp]
            else:
                self.thresholds = threshold_sp
            assert self.thresholds.size == self.grid.number_of_nodes

        # make storage variables
        self.A_to_the_m = grid.zeros(at='node')
        self.alpha = grid.empty(at='node')

        if self.K is None:
            raise ValueError('K_sp must be set as a float, node array, or ' +
                             'field name. It was None.')

        # now handle the inputs that could be float, array or field name:
        # some support here for old-style inputs
        if isinstance(K_sp, string_types):
            if K_sp == 'array':
                self.K = None
            else:
                self.K = self._grid.at_node[K_sp]
        elif isinstance(K_sp, (float, int)):
            self.K = float(K_sp)
        else:
            self.K = np.asarray(K_sp, dtype=float)
            if len(self.K) != self.grid.number_of_nodes:
                raise TypeError('Supplied value of K_sp is not n_nodes long')

        if isinstance(rainfall_intensity, string_types):
            raise ValueError('This component can no longer handle ' +
                             'spatially variable runoff directly. Use ' +
                             'FlowAccumulator with specified ' +
                             'water__unit_flux_in, or use StreamPowerEroder' +
                             'component instead of FastscapeEroder.')
            if rainfall_intensity == 'array':
                self._r_i = None
            else:
                self._r_i = self._grid.at_node[rainfall_intensity]
        elif isinstance(rainfall_intensity, (float, int)):  # a float
            self._r_i = float(rainfall_intensity)
        elif len(rainfall_intensity) == self.grid.number_of_nodes:
            raise ValueError('This component can no longer handle '
                             'spatially variable runoff directly. Use '
                             'FlowAccumulator with specified '
                             'water__unit_flux_in, or use StreamPowerEroder'
                             'component instead of FastscapeEroder.')
            self._r_i = np.array(rainfall_intensity)
        else:
            raise TypeError('Supplied type of rainfall_intensity was '
                            'not recognised')

        # We now forbid changing of the field name
        if 'value_field' in kwds.keys():
            raise ValueError('This component can no longer support variable'
                             'field names. Use "topographic__elevation".')

        # Handle option for area vs discharge
        self.discharge_name = discharge_name

    def erode(self, grid_in, dt=None, K_if_used=None, flooded_nodes=None,
              rainfall_intensity_if_used=None):
        """Erode using stream power erosion.

        This method implements the stream power erosion, following the Braun-
        Willett (2013) implicit Fastscape algorithm. This should allow it to
        be stable against larger timesteps than an explicit stream power
        scheme.

        This driving method for this component is now superceded by the new,
        standardized wrapper :func:`run_one_step`, but is retained for
        back compatibility.

        Set *K_if_used* as a field name or nnodes-long array if you set
        *K_sp* as *"array"* during initialization.

        It returns the grid, in which it will have modified the value of
        *value_field*, as specified in component initialization.

        Parameters
        ----------
        grid_in : a grid
            This is a dummy argument maintained for component back-
            compatibility. It is superceded by the copy of the grid passed
            during initialization.
        dt : float
            Time-step size. If you are calling the deprecated function
            :func:`gear_timestep`, that method will supercede any value
            supplied here.
        K_if_used : array (optional)
            Set this to an array if you set K_sp to 'array' in your input file.
        flooded_nodes : ndarray of int (optional)
            IDs of nodes that are flooded and should have no erosion. If not
            provided but flow has still been routed across depressions, erosion
            may still occur beneath the apparent water level (though will
            always still be positive).
        rainfall_intensity_if_used : float or None (optional)
            Supply to drive this component with a time-varying spatially
            constant rainfall.

        Returns
        -------
        grid
            A reference to the grid.
        """
        upstream_order_IDs = self._grid.at_node['flow__upstream_node_order']
        z = self._grid.at_node['topographic__elevation']
        defined_flow_receivers = np.not_equal(
            self._grid.at_node['flow__link_to_receiver_node'], UNDEFINED_INDEX)

        if isinstance(self._grid, RasterModelGrid):
            flow_link_lengths = self._grid.length_of_d8[
                self._grid.at_node['flow__link_to_receiver_node'][
                    defined_flow_receivers]]
        else:
            flow_link_lengths = self._grid.length_of_link[
                self._grid.at_node['flow__link_to_receiver_node'][
                    defined_flow_receivers]]
        # make arrays from input the right size
        if isinstance(self.K, np.ndarray):
            K_here = self.K[defined_flow_receivers]
        else:
            K_here = self.K
        if rainfall_intensity_if_used is not None:
            assert type(rainfall_intensity_if_used) in (float, np.float64,
                                                        int)
            r_i_here = float(rainfall_intensity_if_used)
        else:
            r_i_here = self._r_i

        if dt is None:
            dt = self.dt
        assert dt is not None, ('Fastscape component could not find a dt to ' +
                                'use. Pass dt to the run_one_step() method.')

        if self.K is None:  # "old style" setting of array
            assert K_if_used is not None
            self.K = K_if_used

        n = float(self.n)

        np.power(self._grid['node'][self.discharge_name], self.m,
                 out=self.A_to_the_m)
        self.alpha[defined_flow_receivers] = (
            r_i_here**self.m * K_here * dt * self.A_to_the_m[
                defined_flow_receivers] / (flow_link_lengths**self.n))

        flow_receivers = self._grid['node']['flow__receiver_node']
        alpha = self.alpha

        # Handle flooded nodes, if any (no erosion there)
        if flooded_nodes is not None:
            alpha[flooded_nodes] = 0.
        else:
            reversed_flow = z < z[flow_receivers]
            # this check necessary if flow has been routed across depressions
            alpha[reversed_flow] = 0.

        threshsdt = self.thresholds * dt

        # solve using Brent's Method in Cython for Speed
        if isinstance(self.thresholds, float):
            brent_method_erode_fixed_threshold(
                upstream_order_IDs, flow_receivers, threshsdt, alpha, n, z)
        else:
            brent_method_erode_variable_threshold(
                upstream_order_IDs, flow_receivers, threshsdt, alpha, n, z)

        return self._grid

    def run_one_step(self, dt, flooded_nodes=None,
                     rainfall_intensity_if_used=None, **kwds):
        """Erode for a single time step.

        This method implements the stream power erosion across one time
        interval, dt, following the Braun-Willett (2013) implicit Fastscape
        algorithm.

        This follows Landlab standardized component design, and supercedes the
        old driving method :func:`erode`.

        Parameters
        ----------
        dt : float
            Time-step size
        flooded_nodes : ndarray of int (optional)
            IDs of nodes that are flooded and should have no erosion. If not
            provided but flow has still been routed across depressions, erosion
            may still occur beneath the apparent water level (though will
            always still be positive).
        rainfall_intensity_if_used : float or None (optional)
            Supply to drive this component with a time-varying spatially
            constant rainfall.
        """
        self.erode(grid_in=self._grid, dt=dt, flooded_nodes=flooded_nodes,
                   rainfall_intensity_if_used=rainfall_intensity_if_used)
